"""Black & Litterman Prior Model estimator."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# Riskfolio-Lib, Copyright (c) 2020-2023, Dany Cajas, Licensed under BSD 3 clause.
# PyPortfolioOpt, Copyright (c) 2018 Robert Andrew Martin, Licensed under MIT Licence.

import numpy as np
import numpy.typing as npt

from skfolio.moments import EquilibriumMu
from skfolio.prior._base import BasePrior, PriorModel
from skfolio.prior._empirical import EmpiricalPrior
from skfolio.utils.equations import equations_to_matrix
from skfolio.utils.tools import check_estimator, input_to_array


class BlackLitterman(BasePrior):
    """Black & Litterman Prior Model estimator.

    The Black & Litterman model [1]_ takes a Bayesian approach by using a prior estimate
    of the assets expected returns and covariance matrix, which are updated using the
    analyst views to get a posterior estimate.

    Parameters
    ----------
    views : array-like of floats of shape (n_views,)
        The analyst views about the assets expected returns.
        The views must match the following patterns:

            * Absolute view: "asset_i = a"
            * Relative view: "asset_i - asset_j = b"

        With "asset_i" and "asset_j" the assets names and "a" and "b" the analyst views
        about the assets expected returns expressed in the same frequency as the
        returns `X`.

        Examples:

            * "SPX = 0.00015" --> the SPX will have a daily expected return of 0.015%
            * "SX5E - TLT = 0.00039" --> the SX5E will outperform the TLT by a daily expected return of 0.039%
            * "SX5E - SPX = -0.0002" --> the SX5E will underperform the SPX by a daily expected return of 0.02%
            * "Equity = 0.00010" --> the sum of Equity assets will have a daily expected return of 0.01%
            * "Europe - US = 0.0004" --> the sum of European assets will outperform the sum of US assets by a daily expected return of 0.04%

    groups : dict[str, list[str]] or array-like of strings of shape (n_groups, n_assets), optional
        The assets groups to be referenced in `views`.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset groups) and the input `X` of the `fit` method must be a
        DataFrame with the assets names in columns.

        Examples:

            * groups = {"SX5E": ["Equity", "Europe"], "SPX": ["Equity", "US"], "TLT": ["Bond", "US"]}
            * groups = [["Equity", "Equity", "Bond"], ["Europe", "US", "US"]]

    prior_estimator : BasePrior, optional
        The assets' :ref:`prior model estimator <prior>`. It is used to estimate
        the :class:`~skfolio.prior.PriorModel` containing the estimation of the assets
        expected returns, covariance matrix, returns and Cholesky decomposition.
        The default (`None`) is to use `EmpiricalPrior(mu_estimator=EquilibriumMu())`.

    tau : float, default=0.05
        Tau controls the degree of uncertainty given to the analyst views. A low value
        means high uncertainty and will put less weight on the analyst views compared to
        the prior returns. The default value is `0.05`.
        Other common values used in the literature are `1.0` or the inverse of the
        number of observations.

    view_confidences : array-like of floats of shape (n_views,), optional
        Instead of using a diagonal uncertainty matrix (Omega) proportional to the prior
        covariance matrix, you can provide the vector of view confidences (between 0
        and 1) as describe by the Idzorek's method [2]_.

    risk_free_rate : float, default=0.0
        The risk-free rate.

    Attributes
    ----------
    prior_model_ : PriorModel
        The :class:`~skfolio.prior.PriorModel`.

    groups_ : ndarray of shape(n_groups, n_assets)
        Assets names and groups converted to an 2D array.

    views_ : ndarray of shape (n_views,)
        The analyst views converted to a ndarray of floats.

    picking_matrix_ : ndarray of shape (n_views, n_assets)
        Picking matrix computed from the views and assets names/groups.

    prior_estimator_ : BasePrior
        Fitted `prior_estimator`.

    n_features_in_ : int
       Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
       Names of features seen during `fit`. Defined only when `X`
       has feature names that are all strings.

    References
    ----------
    .. [1]  "Combining investor views with market equilibrium",
        The Journal of Fixed Income,
        Fischer Black and Robert Litterman, 1991.

    .. [2]  "A step-by-step guide to the Black-Litterman model : Incorporating
        user-specified confidence",
        Forecasting Expected Returns in the Financial Markets,
        Idzorek T, 2007.
    """

    groups_: np.ndarray
    views_: np.ndarray
    picking_matrix_: np.ndarray
    prior_estimator_: BasePrior

    def __init__(
        self,
        views: npt.ArrayLike,
        groups: dict[str, list[str]] | npt.ArrayLike | None = None,
        prior_estimator: BasePrior | None = None,
        tau: float = 0.05,
        view_confidences: npt.ArrayLike | None = None,
        risk_free_rate: float = 0,
    ):
        self.views = views
        self.groups = groups
        self.prior_estimator = prior_estimator
        self.tau = tau
        self.view_confidences = view_confidences
        self.risk_free_rate = risk_free_rate

    def fit(self, X: npt.ArrayLike, y=None) -> "BlackLitterman":
        """Fit the Black & Litterman estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : BlackLitterman
            Fitted estimator.
        """
        self.prior_estimator_ = check_estimator(
            self.prior_estimator,
            default=EmpiricalPrior(mu_estimator=EquilibriumMu()),
            check_type=BasePrior,
        )
        # fitting prior estimator
        self.prior_estimator_.fit(X)

        prior_mu = self.prior_estimator_.prior_model_.mu
        prior_covariance = self.prior_estimator_.prior_model_.covariance
        prior_returns = self.prior_estimator_.prior_model_.returns

        # we validate after all models have been fitted to keep features names
        # information.
        self._validate_data(X)

        n_assets = prior_returns.shape[1]
        views = np.asarray(self.views)
        if views.ndim != 1:
            raise ValueError(f"`views` must be a 1D array, got a {views.ndim}D array.")
        if self.groups is None:
            if not hasattr(self, "feature_names_in_"):
                raise ValueError(
                    "You must provide either `groups`"
                    " or `X` as a DataFrame with asset names in columns"
                )
            self.groups_ = np.asarray([self.feature_names_in_])
        else:
            self.groups_ = input_to_array(
                items=self.groups,
                n_assets=n_assets,
                fill_value="",
                dim=2,
                assets_names=(
                    self.feature_names_in_
                    if hasattr(self, "feature_names_in_")
                    else None
                ),
                name="groups",
            )
        self.picking_matrix_, self.views_ = equations_to_matrix(
            groups=self.groups_,
            equations=views,
            sum_to_one=True,
            raise_if_group_missing=True,
            names=("groups", "views"),
        )

        if self.view_confidences is None:
            omega = np.diag(
                np.diag(
                    self.tau
                    * self.picking_matrix_
                    @ prior_covariance
                    @ self.picking_matrix_.T
                )
            )
        else:
            # Idzorek's method using Jay Walters closed form solution
            view_confidences = np.asarray(self.view_confidences)
            if np.any(view_confidences < 0) or np.any(view_confidences > 1):
                raise ValueError(
                    "all values of view_confidences must be between 0 and 1"
                )
            view_confidences[view_confidences == 0] = 1e-16
            alphas = 1 / view_confidences - 1
            omega = np.diag(
                np.diag(
                    self.tau
                    * alphas[:, np.newaxis]
                    * self.picking_matrix_
                    @ prior_covariance
                    @ self.picking_matrix_.T
                )
            )

        # solving linear system instead of matrix inversion
        _v = self.tau * prior_covariance @ self.picking_matrix_.T
        _a = self.picking_matrix_ @ _v + omega
        _b = self.views_ - self.picking_matrix_ @ prior_mu
        posterior_mu = prior_mu + _v @ np.linalg.solve(_a, _b) + self.risk_free_rate
        posterior_covariance = (
            prior_covariance
            + self.tau * prior_covariance
            - _v @ np.linalg.solve(_a, _v.T)
        )
        self.prior_model_ = PriorModel(
            mu=posterior_mu, covariance=posterior_covariance, returns=prior_returns
        )
        return self



"""Hierarchical Clustering estimators."""

# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

from skfolio.cluster._hierarchical import HierarchicalClustering, LinkageMethod

__all__ = ["LinkageMethod", "HierarchicalClustering"]



"""
The :mod:`skfolio.exceptions` module includes all custom warnings and error
classes used across skfolio.
"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

__all__ = [
    "OptimizationError",
    "EquationToMatrixError",
    "GroupNotFoundError",
    "NonPositiveVarianceError",
]


class OptimizationError(Exception):
    """Optimization Did not converge"""


class EquationToMatrixError(Exception):
    """Error while processing equations"""


class GroupNotFoundError(Exception):
    """Group name not found in the groups"""


class NonPositiveVarianceError(Exception):
    """Variance negative or null"""



"""Base Prior estimator"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

from abc import ABC, abstractmethod
from dataclasses import dataclass

import numpy as np
import numpy.typing as npt
import sklearn.base as skb


# frozen=True with eq=False will lead to an id-based hashing which is needed for
# caching CVX models in Optimization without impacting performance
@dataclass(frozen=True, eq=False)
class PriorModel:
    """Prior model dataclass.

    Attributes
    ----------
    mu : ndarray of shape (n_assets,)
        Estimation of the assets expected returns.

    covariance : ndarray of shape (n_assets, n_assets)
        Estimation of the assets covariance matrix.

    returns : ndarray of shape (n_observations, n_assets)
        Estimation of the assets returns.

    cholesky : ndarray, optional
        Lower-triangular Cholesky factor of the covariance. In some cases it is possible
        to obtain a cholesky factor with less dimension compared to the one obtained
        directly by applying the cholesky decomposition to the covariance estimation
        (for example in Factor Models). When provided, this cholesky factor is use in
        some optimizations (for example in mean-variance) to improve performance and
        convergence. The default is `None`.
    """

    mu: np.ndarray
    covariance: np.ndarray
    returns: np.ndarray
    cholesky: np.ndarray | None = None


class BasePrior(skb.BaseEstimator, ABC):
    """Base class for all prior estimators in skfolio.

    Notes
    -----
    All estimators should specify all the parameters that can be set
    at the class level in their ``__init__`` as explicit keyword
    arguments (no ``*args`` or ``**kwargs``).
    """

    prior_model_: PriorModel

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def fit(self, X: npt.ArrayLike, y=None):
        pass



"""Risk Budgeting Optimization estimator."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# The optimization features are derived
# from Riskfolio-Lib, Copyright (c) 2020-2023, Dany Cajas, Licensed under BSD 3 clause.

import cvxpy as cp
import numpy as np
import numpy.typing as npt

import skfolio.typing as skt
from skfolio.measures import RiskMeasure
from skfolio.optimization.convex._base import ConvexOptimization
from skfolio.prior import BasePrior, EmpiricalPrior
from skfolio.utils.tools import args_names, check_estimator


class RiskBudgeting(ConvexOptimization):
    r"""Risk Budgeting Optimization estimator.

    The Risk Budgeting estimator solves the below convex problem:

        .. math::   \begin{cases}
                    \begin{aligned}
                    &\min_{w} & & risk_{i}(w) \\
                    &\text{s.t.} & & b^T \cdot log(w) \ge c \\
                    & & & w^T \cdot \mu \ge min\_return \\
                    & & & A \cdot w \ge b \\
                    & & & w \ge 0
                    \end{aligned}
                    \end{cases}

    with :math:`b` the risk budget vector and :math:`c` an auxiliary variable of
    the log barrier.

    And :math:`risk_{i}` a risk measure among:

        * Mean Absolute Deviation
        * First Lower Partial Moment
        * Variance
        * Semi-Variance
        * CVaR (Conditional Value at Risk)
        * EVaR (Entropic Value at Risk)
        * Worst Realization (worst return)
        * CDaR (Conditional Drawdown at Risk)
        * Maximum Drawdown
        * Average Drawdown
        * EDaR (Entropic Drawdown at Risk)
        * Ulcer Index
        * Gini Mean Difference

    Cost and additional constraints can also be added to the optimization problem  (see
    the parameters description).

    Limitations are imposed on some constraints including long only weights to ensure
    convexity.

    The assets expected returns, covariance matrix and returns are estimated from the
    :ref:`prior estimator <prior>`.

    Parameters
    ----------
    risk_measure : RiskMeasure, default=RiskMeasure.VARIANCE
        :class:`~skfolio.meta.RiskMeasure` of the optimization.
        Can be any of:

            * VARIANCE
            * SEMI_VARIANCE
            * STANDARD_DEVIATION
            * SEMI_DEVIATION
            * MEAN_ABSOLUTE_DEVIATION
            * FIRST_LOWER_PARTIAL_MOMENT
            * CVAR
            * EVAR
            * WORST_REALIZATION
            * CDAR
            * MAX_DRAWDOWN
            * AVERAGE_DRAWDOWN
            * EDAR
            * ULCER_INDEX
            * GINI_MEAN_DIFFERENCE_RATIO

        The default is `RiskMeasure.VARIANCE`.

    risk_budget : dict[str, float] | array-like of shape (n_assets,), optional
        Risk budget allocated to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset risk budget) and the input `X` of the `fit` method must be a
        DataFrame with the assets names in columns.
        The default (`None`) is to use the identity vector, reducing the risk
        budgeting to a risk-parity (each asset contributing equally to the total risk).

    prior_estimator : BasePrior, optional
        :ref:`Prior estimator <prior>`.
        The prior estimator is used to estimate the :class:`~skfolio.prior.PriorModel`
        containing the estimation of assets expected returns, covariance matrix,
        returns and Cholesky decomposition of the covariance.
        The default (`None`) is to use :class:`~skfolio.prior.EmpiricalPrior`.

    min_weights : float | dict[str, float] | array-like of shape (n_assets, ) | None, default=0.0
        Minimum assets weights (weights lower bounds).
        If a float is provided, it is applied to each asset.
        `None` is equivalent to `-np.Inf` (no lower bound).
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset minium weight) and the input `X` of the `fit` method must
        be a DataFrame with the assets names in columns.
        When using a dictionary, assets values that are not provided are assigned
        a minimum weight of `0.0`.
        The default value is `0.0` (no short selling).

        Example:

           * `min_weights = 0` --> long only portfolio (no short selling).
           * `min_weights = None` --> no lower bound (same as `-np.Inf`).
           * `min_weights = -2` --> each weight must be above -200%.
           * `min_weights = {"SX5E": 0, "SPX": -2}`
           * `min_weights = [0, -2]`

    max_weights : float | dict[str, float] | array-like of shape (n_assets, ) | None, default=1.0
        Maximum assets weights (weights upper bounds).
        If a float is provided, it is applied to each asset.
        `None` is equivalent to `+np.Inf` (no upper bound).
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset maximum weight) and the input `X` of the `fit` method must
        be a DataFrame with the assets names in columns.
        When using a dictionary, assets values that are not provided are assigned
        a minimum weight of `1.0`.
        The default value is `1.0` (each asset is below 100%).

        Example:

           * `max_weights = 0` --> no long position (short only portfolio).
           * `max_weights = None` --> no upper bound.
           * `max_weights = 2` --> each weight must be below 200%.
           * `max_weights = {"SX5E": 1, "SPX": 2}`
           * `max_weights = [1, 2]`

    transaction_costs : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Transaction costs of the assets. It is used to add linear transaction costs to
        the optimization problem:

        .. math:: total\_cost = \sum_{i=1}^{N} c_{i} \times |w_{i} - w\_prev_{i}|

        with :math:`c_{i}` the transaction cost of asset i, :math:`w_{i}` its weight
        and :math:`w\_prev_{i}` its previous weight (defined in `previous_weights`).
        The float :math:`total\_cost` is impacting the portfolio expected return in the optimization:

        .. math:: expected\_return = \mu^{T} \cdot w - total\_cost

        with :math:`\mu` the vector af assets' expected returns and :math:`w` the
        vector of assets weights.

        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset cost) and the input `X` of the `fit` method must be a
        DataFrame with the assets names in columns.
        The default value is `0.0`.

        .. warning::

            Based on the above formula, the periodicity of the transaction costs
            needs to be homogenous to the periodicity of :math:`\mu`. For example, if
            the input `X` is composed of **daily** returns, the `transaction_costs` need
            to be expressed as **daily** costs.
            (See :ref:`sphx_glr_auto_examples_1_mean_risk_plot_6_transaction_costs.py`)

    management_fees : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Management fees of the assets. It is used to add linear management fees to the
        optimization problem:

        .. math:: total\_fee = \sum_{i=1}^{N} f_{i} \times w_{i}

        with :math:`f_{i}` the management fee of asset i and :math:`w_{i}` its weight.
        The float :math:`total\_fee` is impacting the portfolio expected return in the optimization:

        .. math:: expected\_return = \mu^{T} \cdot w - total\_fee

        with :math:`\mu` the vector af assets expected returns and :math:`w` the vector
        of assets weights.

        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset fee) and the input `X` of the `fit` method must be a
        DataFrame with the assets names in columns.
        The default value is `0.0`.

        .. warning::

            Based on the above formula, the periodicity of the management fees needs to
            be homogenous to the periodicity of :math:`\mu`. For example, if the input
            `X` is composed of **daily** returns, the `management_fees` need to be
            expressed in **daily** fees.

        .. note::

            Another approach is to directly impact the management fees to the input `X`
            in order to express the returns net of fees. However, when estimating the
            :math:`\mu` parameter using for example Shrinkage estimators, this approach
            would mix a deterministic value with an uncertain one leading to unwanted
            bias in the management fees.

    previous_weights : float | dict[str, float] | array-like of shape (n_assets, ), optional
        Previous weights of the assets. Previous weights are used to compute the
        portfolio cost and the portfolio turnover.
        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset previous weight) and the input `X` of the `fit` method must
        be a DataFrame with the assets names in columns.
        The default (`None`) means no previous weights.

    linear_constraints : array-like of shape (n_constraints,), optional
       Linear constraints.
        The linear constraints must match any of following patterns:

           * "2.5 * ref1 + 0.10 * ref2 + 0.0013 <= 2.5 * ref3"
           * "ref1 >= 2.9 * ref2"
           * "ref1 <= ref2"
           * "ref1 >= ref1"

        With "ref1", "ref2" ... the assets names or the groups names provided
        in the parameter `groups`. Assets names can be referenced without the need of
        `groups` if the input `X` of the `fit` method is a DataFrame with these
        assets names in columns.

        Examples:

            * "SPX >= 0.10" --> SPX weight must be greater than 10% (note that you can also use `min_weights`)
            * "SX5E + TLT >= 0.2" --> the sum of SX5E and TLT weights must be greater than 20%
            * "US >= 0.7" --> the sum of all US weights must be greater than 70%
            * "Equity <= 3 * Bond" --> the sum of all Equity weights must be less or equal to 3 times the sum of all Bond weights.
            * "2*SPX + 3*Europe <= Bond + 0.05" --> mixing assets and group constraints

    groups : dict[str, list[str]] or array-like of shape (n_groups, n_assets), optional
        The assets groups referenced in `linear_constraints`.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset groups) and the input `X` of the `fit` method must be a
        DataFrame with the assets names in columns.

        Examples:

            * groups = {"SX5E": ["Equity", "Europe"], "SPX": ["Equity", "US"], "TLT": ["Bond", "US"]}
            * groups = [["Equity", "Equity", "Bond"], ["Europe", "US", "US"]]

    left_inequality : array-like of shape (n_constraints, n_assets), optional
        Left inequality matrix :math:`A` of the linear
        constraint :math:`A \cdot w \leq b`.

    right_inequality : array-like of shape (n_constraints, ), optional
        Right inequality vector :math:`b` of the linear
        constraint :math:`A \cdot w \leq b`.

    risk_free_rate : float, default=0.0
        Risk-free interest rate.
        The default value is `0.0`.

    min_return : float | array-like of shape (n_optimization), optional
        Lower bound constraint on the expected return.

    min_acceptable_return : float, optional
        The minimum acceptable return used to distinguish "downside" and "upside"
        returns for the computation of lower partial moments:

            * First Lower Partial Moment
            * Semi-Variance
            * Semi-Deviation

        The default (`None`) is to use the mean.

    cvar_beta : float, default=0.95
        CVaR (Conditional Value at Risk) confidence level.

    evar_beta : float, default=0
        EVaR (Entropic Value at Risk) confidence level.

    cvar_beta : float, default=0.95
        CVaR (Conditional Value at Risk) confidence level.
        The default value is `0.95`.

    evar_beta : float, default=0
        EVaR (Entropic Value at Risk) confidence level.
        The default value is `0.95`.

    cdar_beta : float, default=0.95
        CDaR (Conditional Drawdown at Risk) confidence level.
        The default value is `0.95`.

    edar_beta : float, default=0.95
        EDaR (Entropic Drawdown at Risk) confidence level.
        The default value is `0.95`.

    add_objective : Callable[[cp.Variable], cp.Expression], optional
        Add a custom objective to the existing objective expression.
        It is a function that must take as argument the weights `w` and returns a
        CVXPY expression.

    add_constraints : Callable[[cp.Variable], cp.Expression|list[cp.Expression]], optional
        Add a custom constraint or a list of constraints to the existing constraints.
        It is a function that must take as argument the weights `w` and returns a
        CVPXY expression or a list of CVPXY expressions.

    overwrite_expected_return : Callable[[cp.Variable], cp.Expression], optional
        Overwrite the expected return :math:`\mu \cdot w` with a custom expression.
        It is a function that must take as argument the weights `w` and returns a
        CVPXY expression.

    solver : str, default="CLARABEL"
        The solver to use. The default is "CLARABEL" which is written in Rust and has
        better numerical stability and performance than ECOS and SCS. Cvxpy will replace
        its default solver "ECOS" by "CLARABEL" in future releases.
        For more details about available solvers, check the CVXPY documentation:
        https://www.cvxpy.org/tutorial/advanced/index.html#choosing-a-solver

    solver_params : dict, optional
        Solver parameters. For example, `solver_params=dict(verbose=True)`.
        The default (`None`) is use `{"tol_gap_abs": 1e-9, "tol_gap_rel": 1e-9}`
        for the solver "CLARABEL" and the CVXPY default otherwise.
        For more details about solver arguments, check the CVXPY documentation:
        https://www.cvxpy.org/tutorial/advanced/index.html#setting-solver-options

    scale_objective : float, optional
        Scale each objective element by this value.
        It can be used to increase the optimization accuracies in specific cases.
        The default (`None`) is set depending on the problem.

    scale_constraints : float, optional
        Scale each constraint element by this value.
        It can be used to increase the optimization accuracies in specific cases.
        The default (`None`) is set depending on the problem.

    save_problem : bool, default=False
        If this is set to True, the CVXPY Problem is saved in `problem_`.
        The default is `False`.

    raise_on_failure : bool, default=True
        If this is set to True, an error is raised when the optimization fail otherwise
        it passes with a warning.

    portfolio_params :  dict, optional
        Portfolio parameters passed to the portfolio evaluated by the `predict` and
        `score` methods. If not provided, the `name`, `transaction_costs`,
        `management_fees`, `previous_weights` and `risk_free_rate` are copied from the 
        optimization model and passed to the portfolio.

    Attributes
    ----------
    weights_ : ndarray of shape (n_assets,) or (n_optimizations, n_assets)
        Weights of the assets.

    problem_values_ :  dict[str, float] | list[dict[str, float]] of size n_optimizations
        Expression values retrieved from the CVXPY problem.

    prior_estimator_ : BasePrior
        Fitted `prior_estimator`.

    problem_: cvxpy.Problem
        CVXPY problem used for the optimization. Only when `save_problem` is set to
        `True`.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.
    """

    def __init__(
        self,
        risk_measure: RiskMeasure = RiskMeasure.VARIANCE,
        risk_budget: np.ndarray | None = None,
        prior_estimator: BasePrior | None = None,
        min_weights: skt.MultiInput | None = 0.0,
        max_weights: skt.MultiInput | None = 1.0,
        transaction_costs: skt.MultiInput = 0.0,
        management_fees: skt.MultiInput = 0.0,
        previous_weights: skt.MultiInput | None = None,
        groups: skt.Groups | None = None,
        linear_constraints: skt.LinearConstraints | None = None,
        left_inequality: skt.Inequality | None = None,
        right_inequality: skt.Inequality | None = None,
        risk_free_rate: float = 0.0,
        min_return: skt.Target | None = None,
        min_acceptable_return: skt.Target | None = None,
        cvar_beta: float = 0.95,
        evar_beta: float = 0.95,
        cdar_beta: float = 0.95,
        edar_beta: float = 0.95,
        solver: str = "CLARABEL",
        solver_params: dict | None = None,
        scale_objective: float | None = None,
        scale_constraints: float | None = None,
        save_problem: bool = False,
        raise_on_failure: bool = True,
        add_objective: skt.ExpressionFunction | None = None,
        add_constraints: skt.ExpressionFunction | None = None,
        overwrite_expected_return: skt.ExpressionFunction | None = None,
        portfolio_params: dict | None = None,
    ):
        super().__init__(
            risk_measure=risk_measure,
            prior_estimator=prior_estimator,
            min_weights=min_weights,
            max_weights=max_weights,
            budget=1,
            transaction_costs=transaction_costs,
            management_fees=management_fees,
            previous_weights=previous_weights,
            groups=groups,
            linear_constraints=linear_constraints,
            left_inequality=left_inequality,
            right_inequality=right_inequality,
            risk_free_rate=risk_free_rate,
            min_acceptable_return=min_acceptable_return,
            cvar_beta=cvar_beta,
            evar_beta=evar_beta,
            cdar_beta=cdar_beta,
            edar_beta=edar_beta,
            solver=solver,
            solver_params=solver_params,
            scale_objective=scale_objective,
            scale_constraints=scale_constraints,
            save_problem=save_problem,
            raise_on_failure=raise_on_failure,
            add_objective=add_objective,
            add_constraints=add_constraints,
            overwrite_expected_return=overwrite_expected_return,
            portfolio_params=portfolio_params,
        )
        self.min_return = min_return
        self.risk_budget = risk_budget

    def _validation(self) -> None:
        if not isinstance(self.risk_measure, RiskMeasure):
            raise TypeError("risk_measure must be of type `RiskMeasure`")
        if self.min_weights < 0:
            raise ValueError(
                "Risk Budgeting must have non negative `min_weights` constraint"
                " otherwise the problem becomes non-convex."
            )

    def fit(self, X: npt.ArrayLike, y=None) -> "RiskBudgeting":
        """Fit the Risk Budgeting Optimization estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
           Price returns of the assets.

        y : array-like of shape (n_observations, n_factors), optional
            Price returns of factors.
            The default is `None`.


        Returns
        -------
        self : RiskBudgeting
           Fitted estimator.
        """
        self._check_feature_names(X, reset=True)
        # Validate
        self._validation()
        # Used to avoid adding multiple times similar constrains linked to identical
        # risk models
        self.prior_estimator_ = check_estimator(
            self.prior_estimator,
            default=EmpiricalPrior(),
            check_type=BasePrior,
        )
        self.prior_estimator_.fit(X, y)
        prior_model = self.prior_estimator_.prior_model_
        n_observations, n_assets = prior_model.returns.shape

        # set solvers params
        if self.solver == "CLARABEL":
            self._set_solver_params(default={"tol_gap_abs": 1e-9, "tol_gap_rel": 1e-9})
        else:
            self._set_solver_params(default=None)

        # set scale
        self._set_scale_objective(default=1)
        self._set_scale_constraints(default=1)

        # Risk budget
        risk_budget = self.risk_budget
        if risk_budget is None:
            risk_budget = np.ones(n_assets)
        else:
            risk_budget = self._clean_input(
                self.risk_budget,
                n_assets=n_assets,
                fill_value=1e-10,
                name="risk_budget",
            )
            risk_budget[risk_budget == 0] = 1e-10

        # Variables
        w = cp.Variable(n_assets)
        factor = cp.Variable()
        c = cp.Variable(nonneg=True)

        # Expected returns
        expected_return = (
            self._cvx_expected_return(prior_model=prior_model, w=w)
            - self._cvx_transaction_cost(prior_model=prior_model, w=w, factor=factor)
            - self._cvx_management_fee(prior_model=prior_model, w=w)
        )

        # risk budgeting constraint
        constraints = [
            risk_budget @ cp.log(w) * self._scale_constraints
            >= c * self._scale_constraints
        ]

        # weight constraints
        constraints += self._get_weight_constraints(
            n_assets=n_assets, w=w, factor=factor
        )

        parameters_values = []

        # min_return constraint
        if self.min_return is not None:
            parameter = cp.Parameter(nonneg=False)
            constraints += [
                expected_return * self._scale_constraints
                >= parameter * factor * self._scale_constraints
            ]
            parameters_values.append((parameter, self.min_return))

        # risk and risk constraints
        risk_func = getattr(self, f"_{self.risk_measure.value}_risk")
        args = {}
        for arg_name in args_names(risk_func):
            if arg_name == "prior_model":
                args[arg_name] = prior_model
            elif arg_name == "w":
                args[arg_name] = w
            elif arg_name == "factor":
                if self.risk_measure in [RiskMeasure.FIRST_LOWER_PARTIAL_MOMENT]:
                    args[arg_name] = factor
                else:
                    args[arg_name] = cp.Constant(1)
            else:
                args[arg_name] = getattr(self, arg_name)
        risk, constraints_i = risk_func(**args)
        constraints += constraints_i

        # custom objectives and constraints
        custom_objective = self._get_custom_objective(w=w)
        constraints += self._get_custom_constraints(w=w)

        objective = cp.Minimize(
            risk * self._scale_objective + custom_objective * self._scale_objective
        )

        # problem
        # noinspection PyTypeChecker
        problem = cp.Problem(objective, constraints)

        # results
        self._solve_problem(
            problem=problem,
            w=w,
            factor=factor,
            parameters_values=parameters_values,
            expressions={
                "expected_return": expected_return,
                "risk": risk,
                "factor": factor,
            },
        )

        return self



"""Base Convex Optimization estimator."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# The optimization features are derived
# from Riskfolio-Lib, Copyright (c) 2020-2023, Dany Cajas, Licensed under BSD 3 clause.

import warnings
from abc import ABC, abstractmethod
from enum import auto
from typing import Any

import cvxpy as cp
import cvxpy.constraints.constraint as cpc
import numpy as np
import numpy.typing as npt
import scipy as sc
import scipy.sparse.linalg as scl

import skfolio.typing as skt
from skfolio.measures import RiskMeasure, owa_gmd_weights
from skfolio.optimization._base import BaseOptimization
from skfolio.prior import BasePrior, PriorModel
from skfolio.uncertainty_set import (
    BaseCovarianceUncertaintySet,
    BaseMuUncertaintySet,
    UncertaintySet,
)
from skfolio.utils.equations import equations_to_matrix
from skfolio.utils.tools import AutoEnum, cache_method, input_to_array

INSTALLED_SOLVERS = cp.installed_solvers()


class ObjectiveFunction(AutoEnum):
    r"""Enumeration of objective functions.

    Attributes
    ----------
    MINIMIZE_RISK : str
        Minimize the risk measure.

    MAXIMIZE_RETURN : str
        Maximize the expected return.

    MAXIMIZE_UTILITY : str
        Maximize the utility  :math:`w^T\mu - \lambda \times risk(w)`.

    MAXIMIZE_RATIO : str
        Maximize the ratio  :math:`\frac{w^T\mu - R_{f}}{risk(w)}`.
    """

    MINIMIZE_RISK = auto()
    MAXIMIZE_RETURN = auto()
    MAXIMIZE_UTILITY = auto()
    MAXIMIZE_RATIO = auto()


class ConvexOptimization(BaseOptimization, ABC):
    r"""Base class for all convex optimization estimators in skfolio.

    All risk measures that have a convex formulation are defined in class methods with
    naming convention: `_{risk_measure}_risk`. That naming convention is used for
    dynamic lookup.

    CVX expressions that are shared among multiple risk measures are cached in a
    dictionary named `_cvx_cache`.
    This is to avoid cvx expression duplication and improve performance and convergence.

    Parameters
    ----------
    risk_measure : RiskMeasure, default=RiskMeasure.VARIANCE
        :class:`~skfolio.meta.RiskMeasure` of the optimization.
        Can be any of:

            * VARIANCE
            * SEMI_VARIANCE
            * STANDARD_DEVIATION
            * SEMI_DEVIATION
            * MEAN_ABSOLUTE_DEVIATION
            * FIRST_LOWER_PARTIAL_MOMENT
            * CVAR
            * EVAR
            * WORST_REALIZATION
            * CDAR
            * MAX_DRAWDOWN
            * AVERAGE_DRAWDOWN
            * EDAR
            * ULCER_INDEX
            * GINI_MEAN_DIFFERENCE_RATIO

        The default is `RiskMeasure.VARIANCE`.

    prior_estimator : BasePrior, optional
        :ref:`Prior estimator <prior>`.
        The prior estimator is used to estimate the :class:`~skfolio.prior.PriorModel`
        containing the estimation of assets expected returns, covariance matrix,
        returns and Cholesky decomposition of the covariance.
        The default (`None`) is to use :class:`~skfolio.prior.EmpiricalPrior`.

    min_weights : float | dict[str, float] | array-like of shape (n_assets, ) | None, default=0.0
        Minimum assets weights (weights lower bounds).
        If a float is provided, it is applied to each asset.
        `None` is equivalent to `-np.Inf` (no lower bound).
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset minium weight) and the input `X` of the `fit` method must
        be a DataFrame with the assets names in columns.
        When using a dictionary, assets values that are not provided are assigned
        a minimum weight of `0.0`.
        The default value is `0.0` (no short selling).

        Example:

           * `min_weights = 0` --> long only portfolio (no short selling).
           * `min_weights = None` --> no lower bound (same as `-np.Inf`).
           * `min_weights = -2` --> each weight must be above -200%.
           * `min_weights = {"SX5E": 0, "SPX": -2}`
           * `min_weights = [0, -2]`

    max_weights : float | dict[str, float] | array-like of shape (n_assets, ) | None, default=1.0
        Maximum assets weights (weights upper bounds).
        If a float is provided, it is applied to each asset.
        `None` is equivalent to `+np.Inf` (no upper bound).
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset maximum weight) and the input `X` of the `fit` method must
        be a DataFrame with the assets names in columns.
        When using a dictionary, assets values that are not provided are assigned
        a minimum weight of `1.0`.
        The default value is `1.0` (each asset is below 100%).

        Example:

           * `max_weights = 0` --> no long position (short only portfolio).
           * `max_weights = None` --> no upper bound.
           * `max_weights = 2` --> each weight must be below 200%.
           * `max_weights = {"SX5E": 1, "SPX": 2}`
           * `max_weights = [1, 2]`

    budget : float | None, default=1.0
        Investment budget. It is the sum of long positions and short positions (sum of
        all weights). `None` means no budget constraints.
        The default value is `1.0` (fully invested portfolio).

        Examples:

             * `budget = 1` --> fully invested portfolio.
             * `budget = 0` --> market neutral portfolio.
             * `budget = None` --> no constraints on the sum of weights.

    min_budget : float, optional
        Minimum budget. It is the lower bound of the sum of long and short positions
        (sum of all weights). If provided, you must set `budget=None`.
        The default (`None`) means no minimum budget constraint.

    max_budget :  float, optional
        Maximum budget. It is the upper bound of the sum of long and short positions
        (sum of all weights). If provided, you must set `budget=None`.
        The default (`None`) means no maximum budget constraint.

    max_short : float, optional
        Maximum short position. The short position is defined as the sum of negative
        weights (in absolute term).
        The default (`None`) means no maximum short position.

    max_long : float, optional
        Maximum long position. The long position is defined as the sum of positive
        weights.
        The default (`None`) means no maximum long position.

    transaction_costs : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Transaction costs of the assets. It is used to add linear transaction costs to
        the optimization problem:

        .. math:: total\_cost = \sum_{i=1}^{N} c_{i} \times |w_{i} - w\_prev_{i}|

        with :math:`c_{i}` the transaction cost of asset i, :math:`w_{i}` its weight
        and :math:`w\_prev_{i}` its previous weight (defined in `previous_weights`).
        The float :math:`total\_cost` is impacting the portfolio expected return in the optimization:

        .. math:: expected\_return = \mu^{T} \cdot w - total\_cost

        with :math:`\mu` the vector af assets' expected returns and :math:`w` the
        vector of assets weights.

        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset cost) and the input `X` of the `fit` method must be a
        DataFrame with the assets names in columns.
        The default value is `0.0`.

        .. warning::

            Based on the above formula, the periodicity of the transaction costs
            needs to be homogenous to the periodicity of :math:`\mu`. For example, if
            the input `X` is composed of **daily** returns, the `transaction_costs` need
            to be expressed as **daily** costs.
            (See :ref:`sphx_glr_auto_examples_1_mean_risk_plot_6_transaction_costs.py`)

    management_fees : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Management fees of the assets. It is used to add linear management fees to the
        optimization problem:

        .. math:: total\_fee = \sum_{i=1}^{N} f_{i} \times w_{i}

        with :math:`f_{i}` the management fee of asset i and :math:`w_{i}` its weight.
        The float :math:`total\_fee` is impacting the portfolio expected return in the optimization:

        .. math:: expected\_return = \mu^{T} \cdot w - total\_fee

        with :math:`\mu` the vector af assets expected returns and :math:`w` the vector
        of assets weights.

        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset fee) and the input `X` of the `fit` method must be a
        DataFrame with the assets names in columns.
        The default value is `0.0`.

        .. warning::

            Based on the above formula, the periodicity of the management fees needs to
            be homogenous to the periodicity of :math:`\mu`. For example, if the input
            `X` is composed of **daily** returns, the `management_fees` need to be
            expressed in **daily** fees.

        .. note::

            Another approach is to directly impact the management fees to the input `X`
            in order to express the returns net of fees. However, when estimating the
            :math:`\mu` parameter using for example Shrinkage estimators, this approach
            would mix a deterministic value with an uncertain one leading to unwanted
            bias in the management fees.

    previous_weights : float | dict[str, float] | array-like of shape (n_assets, ), optional
        Previous weights of the assets. Previous weights are used to compute the
        portfolio cost and the portfolio turnover.
        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset previous weight) and the input `X` of the `fit` method must
        be a DataFrame with the assets names in columns.
        The default (`None`) means no previous weights.

    l1_coef : float, default=0.0
        L1 regularization coefficient.
        It is used to penalize the objective function by the L1 norm:

        .. math:: l1\_coef \times \Vert w \Vert_{1} = l1\_coef \times \sum_{i=1}^{N} |w_{i}|

        Increasing this coefficient will reduce the number of non-zero weights
        (cardinality). It tends to increase robustness (out-of-sample stability) but
        reduces diversification.
        The default value is `0.0`.

    l2_coef : float, default=0.0
        L2 regularization coefficient.
        It is used to penalize the objective function by the L2 norm:

        .. math:: l2\_coef \times \Vert w \Vert_{2}^{2} = l2\_coef \times \sum_{i=1}^{N} w_{i}^2

        It tends to increase robustness (out-of-sample stability).
        The default value is `0.0`.

    mu_uncertainty_set_estimator : BaseMuUncertaintySet, optional
        :ref:`Mu Uncertainty set estimator <uncertainty_set_estimator>`.
        If provided, the assets expected returns are modelled with an ellipsoidal
        uncertainty set. It is called worst-case optimization and is a class of robust
        optimization. It reduces the instability that arises from the estimation errors
        of the expected returns.
        The worst case portfolio expect return is:

        .. math:: w^T\hat{\mu} - \kappa_{\mu}\lVert S_{\mu}^\frac{1}{2}w\rVert_{2}

        with :math:`\kappa` the size of the ellipsoid (confidence region) and
        :math:`S` its shape.
        The default (`None`) means that no uncertainty set is used.

    covariance_uncertainty_set_estimator : BaseCovarianceUncertaintySet, optional
        :ref:`Covariance Uncertainty set estimator <uncertainty_set_estimator>`.
        If provided, the assets covariance matrix is modelled with an ellipsoidal
        uncertainty set. It is called worst-case optimization and is a class of robust
        optimization. It reduces the instability that arises from the estimation errors
        of the covariance matrix.
        The default (`None`) means that no uncertainty set is used.

    linear_constraints : array-like of shape (n_constraints,), optional
        Linear constraints.
        The linear constraints must match any of following patterns:

           * "2.5 * ref1 + 0.10 * ref2 + 0.0013 <= 2.5 * ref3"
           * "ref1 >= 2.9 * ref2"
           * "ref1 <= ref2"
           * "ref1 >= ref1"

        With "ref1", "ref2" ... the assets names or the groups names provided
        in the parameter `groups`. Assets names can be referenced without the need of
        `groups` if the input `X` of the `fit` method is a DataFrame with these
        assets names in columns.

        Examples:

            * "SPX >= 0.10" --> SPX weight must be greater than 10% (note that you can also use `min_weights`)
            * "SX5E + TLT >= 0.2" --> the sum of SX5E and TLT weights must be greater than 20%
            * "US >= 0.7" --> the sum of all US weights must be greater than 70%
            * "Equity <= 3 * Bond" --> the sum of all Equity weights must be less or equal to 3 times the sum of all Bond weights.
            * "2*SPX + 3*Europe <= Bond + 0.05" --> mixing assets and group constraints

    groups : dict[str, list[str]] or array-like of shape (n_groups, n_assets), optional
        The assets groups referenced in `linear_constraints`.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset groups) and the input `X` of the `fit` method must be a
        DataFrame with the assets names in columns.

        Examples:

            * groups = {"SX5E": ["Equity", "Europe"], "SPX": ["Equity", "US"], "TLT": ["Bond", "US"]}
            * groups = [["Equity", "Equity", "Bond"], ["Europe", "US", "US"]]

    left_inequality : array-like of shape (n_constraints, n_assets), optional
        Left inequality matrix :math:`A` of the linear
        constraint :math:`A \cdot w \leq b`.

    right_inequality : array-like of shape (n_constraints, ), optional
        Right inequality vector :math:`b` of the linear
        constraint :math:`A \cdot w \leq b`.

    risk_free_rate : float, default=0.0
        Risk-free interest rate.
        The default value is `0.0`.

    min_acceptable_return : float, optional
        The minimum acceptable return used to distinguish "downside" and "upside"
        returns for the computation of lower partial moments:

            * First Lower Partial Moment
            * Semi-Variance
            * Semi-Deviation

        The default (`None`) is to use the mean.

    cvar_beta : float, default=0.95
        CVaR (Conditional Value at Risk) confidence level.
        The default value is `0.95`.

    evar_beta : float, default=0
        EVaR (Entropic Value at Risk) confidence level.
        The default value is `0.95`.

    cdar_beta : float, default=0.95
        CDaR (Conditional Drawdown at Risk) confidence level.
        The default value is `0.95`.

    edar_beta : float, default=0.95
        EDaR (Entropic Drawdown at Risk) confidence level.
        The default value is `0.95`.

    add_objective : Callable[[cp.Variable], cp.Expression], optional
        Add a custom objective to the existing objective expression.
        It is a function that must take as argument the weights `w` and returns a
        CVXPY expression.

    add_constraints : Callable[[cp.Variable], cp.Expression|list[cp.Expression]], optional
        Add a custom constraint or a list of constraints to the existing constraints.
        It is a function that must take as argument the weights `w` and returns a
        CVPXY expression or a list of CVPXY expressions.

    overwrite_expected_return : Callable[[cp.Variable], cp.Expression], optional
        Overwrite the expected return :math:`\mu \cdot w` with a custom expression.
        It is a function that must take as argument the weights `w` and returns a
        CVPXY expression.

    solver : str, default="CLARABEL"
        The solver to use. The default is "CLARABEL" which is written in Rust and has
        better numerical stability and performance than ECOS and SCS. Cvxpy will replace
        its default solver "ECOS" by "CLARABEL" in future releases.
        For more details about available solvers, check the CVXPY documentation:
        https://www.cvxpy.org/tutorial/advanced/index.html#choosing-a-solver

    solver_params : dict, optional
        Solver parameters. For example, `solver_params=dict(verbose=True)`.
        The default (`None`) is use `{"tol_gap_abs": 1e-9, "tol_gap_rel": 1e-9}`
        for the solver "CLARABEL" and the CVXPY default otherwise.
        For more details about solver arguments, check the CVXPY documentation:
        https://www.cvxpy.org/tutorial/advanced/index.html#setting-solver-options

    scale_objective : float, optional
        Scale each objective element by this value.
        It can be used to increase the optimization accuracies in specific cases.
        The default (`None`) is set depending on the problem.

    scale_constraints : float, optional
        Scale each constraint element by this value.
        It can be used to increase the optimization accuracies in specific cases.
        The default (`None`) is set depending on the problem.

    save_problem : bool, default=False
        If this is set to True, the CVXPY Problem is saved in `problem_`.
        The default is `False`.

    raise_on_failure : bool, default=True
        If this is set to True, an error is raised when the optimization fail otherwise
        it passes with a warning.

    portfolio_params :  dict, optional
        Portfolio parameters passed to the portfolio evaluated by the `predict` and
        `score` methods. If not provided, the `name`, `transaction_costs`,
        `management_fees`, `previous_weights` and `risk_free_rate` are copied from the
        optimization model and passed to the portfolio.

    Attributes
    ----------
    weights_ : ndarray of shape (n_assets,) or (n_optimizations, n_assets)
        Weights of the assets.

    problem_values_ :  dict[str, float] | list[dict[str, float]] of size n_optimizations
        Expression values retrieved from the CVXPY problem.

    prior_estimator_ : BasePrior
        Fitted `prior_estimator`.

    mu_uncertainty_set_estimator_ : BaseMuUncertaintySet
        Fitted `mu_uncertainty_set_estimator` if provided.

    covariance_uncertainty_set_estimator_ : BaseCovarianceUncertaintySet
        Fitted `covariance_uncertainty_set_estimator` if provided.

    problem_: cvxpy.Problem
        CVXPY problem used for the optimization. Only when `save_problem` is set to
        `True`.
    """

    _solver_params: dict
    _scale_objective: cp.Constant
    _scale_constraints: cp.Constant
    _cvx_cache: dict

    problem_: cp.Problem
    problem_values_: dict[str, float] | list[dict[str, float]]
    prior_estimator_: BasePrior
    mu_uncertainty_set_estimator_: BaseMuUncertaintySet
    covariance_uncertainty_set_estimator_: BaseCovarianceUncertaintySet

    @abstractmethod
    def __init__(
        self,
        risk_measure: RiskMeasure = RiskMeasure.VARIANCE,
        prior_estimator: BasePrior | None = None,
        min_weights: skt.MultiInput | None = 0.0,
        max_weights: skt.MultiInput | None = 1.0,
        budget: float | None = 1.0,
        min_budget: float | None = None,
        max_budget: float | None = None,
        max_short: float | None = None,
        max_long: float | None = None,
        transaction_costs: skt.MultiInput = 0.0,
        management_fees: skt.MultiInput = 0.0,
        previous_weights: skt.MultiInput | None = None,
        groups: skt.Groups | None = None,
        linear_constraints: skt.LinearConstraints | None = None,
        left_inequality: skt.Inequality | None = None,
        right_inequality: skt.Inequality | None = None,
        l1_coef: float = 0.0,
        l2_coef: float = 0.0,
        mu_uncertainty_set_estimator: BaseMuUncertaintySet | None = None,
        covariance_uncertainty_set_estimator: (
            BaseCovarianceUncertaintySet | None
        ) = None,
        risk_free_rate: float = 0.0,
        min_acceptable_return: skt.Target | None = None,
        cvar_beta: float = 0.95,
        evar_beta: float = 0.95,
        cdar_beta: float = 0.95,
        edar_beta: float = 0.95,
        solver: str = "CLARABEL",
        solver_params: dict | None = None,
        scale_objective: float | None = None,
        scale_constraints: float | None = None,
        save_problem: bool = False,
        raise_on_failure: bool = True,
        add_objective: skt.ExpressionFunction | None = None,
        add_constraints: skt.ExpressionFunction | None = None,
        overwrite_expected_return: skt.ExpressionFunction | None = None,
        portfolio_params: dict | None = None,
    ):
        super().__init__(portfolio_params=portfolio_params)
        if risk_measure.is_annualized:
            warnings.warn(
                f"The annualized risk measure {risk_measure} will be converted"
                f"to its non-annualized version {risk_measure.non_annualized_measure}",
                stacklevel=2,
            )
            risk_measure = risk_measure.non_annualized_measure
        self.risk_measure = risk_measure
        self.prior_estimator = prior_estimator
        self.mu_uncertainty_set_estimator = mu_uncertainty_set_estimator
        self.covariance_uncertainty_set_estimator = covariance_uncertainty_set_estimator
        self.min_weights = min_weights
        self.max_weights = max_weights
        self.budget = budget
        self.min_budget = min_budget
        self.max_budget = max_budget
        self.max_short = max_short
        self.max_long = max_long
        self.min_acceptable_return = min_acceptable_return
        self.transaction_costs = transaction_costs
        self.management_fees = management_fees
        self.previous_weights = previous_weights
        self.groups = groups
        self.linear_constraints = linear_constraints
        self.left_inequality = left_inequality
        self.right_inequality = right_inequality
        self.l1_coef = l1_coef
        self.l2_coef = l2_coef
        self.risk_free_rate = risk_free_rate
        self.add_objective = add_objective
        self.add_constraints = add_constraints
        self.overwrite_expected_return = overwrite_expected_return
        self.solver = solver
        self.solver_params = solver_params
        self.save_problem = save_problem
        self.raise_on_failure = raise_on_failure
        self.scale_objective = scale_objective
        self.scale_constraints = scale_constraints
        self.cvar_beta = cvar_beta
        self.evar_beta = evar_beta
        self.cdar_beta = cdar_beta
        self.edar_beta = edar_beta

        self._clear_models_cache()

    def _call_custom_func(
        self, func: skt.ExpressionFunction, w: cp.Variable, name: str = "custom_func"
    ) -> cp.Expression | list[cp.Expression]:
        """Call a user specific function, infer arguments and perform validation.

        Parameters
        ----------
        func : Callable[[cvxpy Variable, any], cvxpy Expression]
            The custom function. Must have one or two positional arguments.
            The first argument is the CVXPY weight variable `w` and the second is
            the reference to the class itself.

        w : cvxpy Variable
            The CVXPY Variable representing assets weights.

        Returns
        -------
        result : cvxpy Expression | list[cvxpy Expression]
            Result of calling the custom function.
        """
        try:
            # noinspection PyUnresolvedReferences
            func_code = func.__code__
        except AttributeError as err:
            raise ValueError("Custom functions is invalid") from err

        if func_code.co_argcount == 1:
            args = (w,)
        elif func_code.co_argcount == 2:
            args = (w, self)
        else:
            raise ValueError(
                "Custom functions must have 1 or 2 positional arguments, got"
                f" {func_code.co_argcount}"
            )
        try:
            return func(*args)
        except Exception as err:
            raise TypeError(
                f"Error while calling {name}. "
                f"{name} must be a function taking as argument "
                "the weight variable OR the weight variable and the estimator object."
            ) from err

    def _clean_input(
        self,
        value: float | dict | npt.ArrayLike | None,
        n_assets: int,
        fill_value: Any,
        name: str,
    ) -> float | np.ndarray:
        """Convert input to cleaned float or ndarray.

        Parameters
        ----------
        value : float, dict, array-like or None.
            Input value to clean.

        n_assets : int
            Number of assets. Used to verify the shape of the converted array.

        fill_value : Any
            When `items` is a dictionary, elements that are not in `asset_names` are
            filled with `fill_value` in the converted array.

        name : str
            Name used for error messages.

        Returns
        -------
        value :  float or ndarray of shape (n_assets,)
            The cleaned float or 1D array.
        """
        if value is None:
            return fill_value
        if np.isscalar(value):
            return float(value)
        return input_to_array(
            items=value,
            n_assets=n_assets,
            fill_value=fill_value,
            dim=1,
            assets_names=(
                self.feature_names_in_ if hasattr(self, "feature_names_in_") else None
            ),
            name=name,
        )

    def _clear_models_cache(self):
        """CLear the cache of CVX models"""
        self._cvx_cache = {}

    def _get_weight_constraints(
        self, n_assets: int, w: cp.Variable, factor: skt.Factor
    ) -> list[cpc.Constraint]:
        """Compute weight constraints from input parameters.

        Parameters
        ----------
        n_assets : int
            Number of assets.

        w : cvxpy Variable
            The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
            Cvxpy variable or constant.

        Returns
        -------
        constrains : list[cvxpy Constrains]
            The list of weights constraints.
        """
        constraints = []

        if self.min_weights is not None:
            min_weights = self._clean_input(
                self.min_weights,
                n_assets=n_assets,
                fill_value=0,
                name="min_weights",
            )
            constraints.append(
                w * self._scale_constraints
                >= min_weights * factor * self._scale_constraints
            )

        if self.max_weights is not None:
            max_weights = self._clean_input(
                self.max_weights,
                n_assets=n_assets,
                fill_value=1,
                name="max_weights",
            )
            constraints.append(
                w * self._scale_constraints
                <= max_weights * factor * self._scale_constraints
            )

        if self.max_long is not None:
            max_long = float(self.max_long)
            if max_long <= 0:
                raise ValueError("`max_long` must be strictly positif")
            constraints.append(
                cp.sum(cp.pos(w)) * self._scale_constraints
                <= max_long * factor * self._scale_constraints
            )

        if self.max_short is not None:
            max_short = float(self.max_short)
            if max_short <= 0:
                raise ValueError("`max_short` must be strictly positif")
            constraints.append(
                cp.sum(cp.neg(w)) * self._scale_constraints
                <= max_short * factor * self._scale_constraints
            )

        if self.min_budget is not None:
            constraints.append(
                cp.sum(w) * self._scale_constraints
                >= float(self.min_budget) * factor * self._scale_constraints
            )

        if self.max_budget is not None:
            constraints.append(
                cp.sum(w) * self._scale_constraints
                <= float(self.max_budget) * factor * self._scale_constraints
            )

        if self.budget is not None:
            if self.max_budget is not None:
                raise ValueError(
                    "`max_budget`and `budget` cannot be provided at the same time"
                )
            if self.min_budget is not None:
                raise ValueError(
                    "`min_budget`and `budget` cannot be provided at the same time"
                )
            constraints.append(
                cp.sum(w) * self._scale_constraints
                == float(self.budget) * factor * self._scale_constraints
            )

        if self.linear_constraints is not None:
            if self.groups is None:
                if not hasattr(self, "feature_names_in_"):
                    raise ValueError(
                        "If `linear_constraints` is provided you must provide either"
                        " `groups` or `X` as a DataFrame with asset names in columns"
                    )
                groups = np.asarray([self.feature_names_in_])
            else:
                groups = input_to_array(
                    items=self.groups,
                    n_assets=n_assets,
                    fill_value="",
                    dim=2,
                    assets_names=(
                        self.feature_names_in_
                        if hasattr(self, "feature_names_in_")
                        else None
                    ),
                    name="groups",
                )
            a, b = equations_to_matrix(
                groups=groups,
                equations=self.linear_constraints,
                raise_if_group_missing=False,
            )
            if np.any(a != 0):
                constraints.append(
                    a @ w * self._scale_constraints
                    - b * factor * self._scale_constraints
                    <= 0
                )

        if self.left_inequality is not None and self.right_inequality is not None:
            left_inequality = np.asarray(self.left_inequality)
            right_inequality = np.asarray(self.right_inequality)
            if left_inequality.ndim != 2:
                raise ValueError(
                    f"`left_inequality` must be a 2D array, got {left_inequality.ndim}D"
                    " array"
                )
            if right_inequality.ndim != 1:
                raise ValueError(
                    "`right_inequality` must be a 1D array, got"
                    f" {right_inequality.ndim}D array"
                )
            if left_inequality.shape[1] != n_assets:
                raise ValueError(
                    "`left_inequality` must be of shape (n_inequalities, n_assets) "
                    f"with n_assets={n_assets}, got {left_inequality.shape[1]}"
                )
            if left_inequality.shape[0] != right_inequality.shape[0]:
                raise ValueError(
                    "`left_inequality` and `right_inequality` must have same number of"
                    f" rows (i.e. n_inequalities) , got {left_inequality.shape[0]} and"
                    f" {right_inequality.shape[0]}"
                )
            constraints.append(
                left_inequality @ w * self._scale_constraints
                - right_inequality * factor * self._scale_constraints
                <= 0
            )

        return constraints

    def _set_solver_params(self, default: dict | None) -> None:
        """Set the solver params by saving its value in `_solver_params`.
        It uses `solver` if provided otherwise it uses the `default` solver.

        Parameters
        ----------
        default : str
            The default solver params to use when `solver_params` is `None`.
        """
        if self.solver_params is None:
            self._solver_params = default if default is not None else {}
        else:
            self._solver_params = self.solver_params

    def _set_scale_objective(self, default: float) -> None:
        """Set the objective scale by saving its value in `_scale_objective`.
        It uses `scale_objective` if provided otherwise it uses the `default` scale.

        Parameters
        ----------
        default : float
            The default objective scale to use when `scale_objective` is `None`.
        """
        if self.scale_objective is None:
            self._scale_objective = cp.Constant(default)
        else:
            self._scale_objective = cp.Constant(self.scale_objective)

    def _set_scale_constraints(self, default: float) -> None:
        """Set the constraints scale by saving its value in `_scale_constraints`.
        It uses `scale_constraints` if provided otherwise it uses the `default` scale.

        Parameters
        ----------
        default : float
            The default constraints scale to use when `scale_constraints` is `None`.
        """
        if self.scale_constraints is None:
            self._scale_constraints = cp.Constant(default)
        else:
            self._scale_constraints = cp.Constant(self.scale_constraints)

    def _get_custom_objective(self, w: cp.Variable) -> cp.Expression:
        """Returns the CVXPY expression evaluated by calling the `add_objective`
        function if provided, otherwise returns the CVXPY constant `0`.

        Parameters
        ----------
        w : cvxpy Variable
            The CVXPY Variable representing assets weights.

        Returns
        -------
        expression : cvxpy Expression
            The CVXPY expression evaluated by calling the `add_objective`
            function if provided, otherwise returns the CVXPY constant `0`.
        """
        if self.add_objective is None:
            return cp.Constant(0)
        return self._call_custom_func(
            func=self.add_objective, w=w, name="add_objective"
        )

    def _get_custom_constraints(self, w: cp.Variable) -> list[cp.Expression]:
        """Returns the list of CVXPY expressions evaluated by calling the
        `add_constraint`s function if provided, otherwise returns an empty list.

        Parameters
        ----------
        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        Returns
        -------
        expressions : list of cvxpy Expression
            The list of CVXPY expressions evaluated by calling the
            `add_constraints` function if provided, otherwise returns an empty list.
        """
        if self.add_constraints is None:
            return []
        constraints = self._call_custom_func(
            func=self.add_constraints, w=w, name="add_constraint"
        )
        if isinstance(constraints, list):
            return constraints
        return [constraints]

    @cache_method("_cvx_cache")
    def _cvx_expected_return(
        self, prior_model: PriorModel, w: cp.Variable
    ) -> cp.Expression:
        """Expected Return expression"""
        if self.overwrite_expected_return is None:
            expected_return = prior_model.mu @ w
        else:
            expected_return = self._call_custom_func(
                func=self.overwrite_expected_return,
                w=w,
                name="overwrite_expected_return",
            )
        return expected_return

    # Model reused among multiple risk measure
    def _solve_problem(
        self,
        problem: cp.Problem,
        w: cp.Variable,
        factor: skt.Factor,
        parameters_values: skt.ParametersValues = None,
        expressions: dict[str, cp.Expression] | None = None,
    ) -> None:
        """Solve the CVXPY Problem and save the results in `weights_`, `problem_values_`
        and `problem_`.

        Parameters
        ----------
        problem : cvxpy Problem
            The CVXPY Problem.

        w : cvxpy Variable
            The CVXPY Variable representing assets weights.

        expressions : dict[str, cvxpy Expression] | None, optional
            Dictionary of CVXPY Expressions from which values are retrieved and saved
            in `expression_values_`. It is used to save additional information about
            the problem.

        parameters_values: list[tuple[cvxpy Parameter, float | ndarray]], optional
            A list of tuple of CVXPY Parameter and their values.
            If The values are ndarray instead of float, the optimization is solved for
            each element in the array.

        factor: cvxpy Variable | cvxpy Constant
           CVXPY Variable or Constant used for RatioMeasure optimization problems.
        """
        if self.solver not in INSTALLED_SOLVERS:
            raise ValueError(f"The solver {self.solver} is not installed.")

        if parameters_values is None:
            parameters_values = []

        if expressions is None:
            expressions = {}

        n_optimizations = 1
        if len(parameters_values) != 0:
            # If the parameter value is a list, each element is the parameter value of
            # a distinct optimization. Therefore, each list must have same length.
            sizes = [len(v) for p, v in parameters_values if not np.isscalar(v)]
            if not np.all(sizes):
                raise ValueError(
                    "All list elements from `parameters_values` should have same length"
                )
            if len(sizes) != 0:
                n_optimizations = sizes[0]
            # Scalar parameter values will be used in each optimization, therefore we
            # transform them to a list.
            parameters_values = [
                (p, [v] * n_optimizations) if np.isscalar(v) else (p, v)
                for p, v in parameters_values
            ]

        all_weights = []
        all_problem_values = []
        optimal = True
        for i in range(n_optimizations):
            for parameter, values in parameters_values:
                parameter.value = values[i]

            try:
                # We suppress cvxpy warning as it is redundant with our warning
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    problem.solve(solver=self.solver, **self._solver_params)

                if w.value is None:
                    raise cp.SolverError("No solution found")

                weights = w.value / factor.value
                problem_values = {
                    name: expression.value / factor.value
                    for name, expression in expressions.items()
                }
                problem_values["objective"] = (
                    problem.value / self._scale_objective.value
                )

                if (
                    self.risk_measure
                    in [RiskMeasure.VARIANCE, RiskMeasure.SEMI_VARIANCE]
                    and "risk" in problem_values
                ):
                    problem_values["risk"] /= factor.value

                all_problem_values.append(problem_values)
                all_weights.append(np.array(weights, dtype=float))

                if problem.status != cp.OPTIMAL:
                    optimal = False
            except (cp.SolverError, scl.ArpackNoConvergence):
                params_string = " ".join(
                    [f"{p.value:0g}" for p in problem.parameters()]
                )
                if len(params_string) != 0:
                    params_string = f" with parameters {params_string}"
                msg = (
                    f"Solver '{self.solver}' failed for {params_string}. Try another"
                    " solver, or solve with solver_params=dict(verbose=True) for more"
                    " information"
                )
                if self.raise_on_failure:
                    raise cp.SolverError(msg) from None
                else:
                    warnings.warn(msg, stacklevel=2)

        if not optimal:
            warnings.warn(
                "Solution may be inaccurate. Try changing the solver params or the"
                " scale. For more details, set `solver_params=dict(verbose=True)`",
                stacklevel=2,
            )

        if n_optimizations == 1:
            self.weights_ = all_weights[0]
            self.problem_values_ = all_problem_values[0]
        else:
            self.weights_ = np.array(all_weights, dtype=float)
            self.problem_values_ = all_problem_values

        if self.save_problem:
            self.problem_ = problem

        self._clear_models_cache()

    @cache_method("_cvx_cache")
    def _cvx_mu_uncertainty_set(
        self, mu_uncertainty_set: UncertaintySet, w: cp.Variable
    ) -> cp.Expression:
        """Uncertainty Set expression of expected returns.

        Parameters
        ----------
        mu_uncertainty_set : UncertaintySet
            The uncertainty set model of expected returns.

        w : cvxpy Variable
            The CVXPY Variable representing assets weights.

        Returns
        -------
        expression : cvxpy Expression
            The CVXPY Expression of the uncertainty set of expected returns.
        """
        return mu_uncertainty_set.k * cp.pnorm(
            sc.linalg.sqrtm(mu_uncertainty_set.sigma) @ w, 2
        )

    @cache_method("_cvx_cache")
    def _cvx_regularization(self, w: cp.Variable) -> cp.Expression:
        """L1 and L2 regularization expression.

        Parameters
        ----------
        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        Returns
        -------
        expression : cvxpy Expression
           The CVXPY Expression of L1 and L2 regularization.
        """
        # Norm L1
        if self.l1_coef is None or self.l1_coef == 0:
            l1_reg = cp.Constant(0)
        else:
            l1_reg = cp.Constant(self.l1_coef) * cp.norm(w, 1)

        # Norm L2
        if self.l2_coef is None or self.l2_coef == 0:
            l2_reg = cp.Constant(0)
        else:
            l2_reg = self.l2_coef * cp.sum_squares(w)
        regularization = l1_reg + l2_reg
        return regularization

    @cache_method("_cvx_cache")
    def _cvx_transaction_cost(
        self, prior_model: PriorModel, w: cp.Variable, factor: skt.Factor
    ) -> cp.Expression:
        """Transaction cost expression.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
            Additional variable used for the optimization of some objective function
            like the ratio maximization.

        Returns
        -------
        expression : cvxpy Expression
           The CVXPY Expression of transaction cost.
        """
        n_assets = prior_model.returns.shape[1]

        transaction_costs = self._clean_input(
            self.transaction_costs,
            n_assets=n_assets,
            fill_value=0,
            name="transaction_costs",
        )
        if np.all(transaction_costs == 0):
            return cp.Constant(0)

        previous_weights = self._clean_input(
            self.previous_weights,
            n_assets=n_assets,
            fill_value=0,
            name="previous_weights",
        )
        if np.isscalar(previous_weights):
            previous_weights *= np.ones(n_assets)

        if np.isscalar(transaction_costs):
            return transaction_costs * cp.norm(previous_weights * factor - w, 1)
        return cp.norm(
            cp.multiply(transaction_costs, (previous_weights * factor - w)),
            1,
        )

    @cache_method("_cvx_cache")
    def _cvx_management_fee(
        self, prior_model: PriorModel, w: cp.Variable
    ) -> cp.Expression:
        """Management fee expression.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        Returns
        -------
        expression : cvxpy Expression
           The CVXPY Expression of management fee .
        """
        n_assets = prior_model.returns.shape[1]

        management_fees = self._clean_input(
            self.management_fees,
            n_assets=n_assets,
            fill_value=0,
            name="management_fees",
        )
        if np.all(management_fees == 0):
            return cp.Constant(0)

        if np.isscalar(management_fees):
            management_fees *= np.ones(n_assets)
        return management_fees @ w

    @cache_method("_cvx_cache")
    def _cvx_returns(self, prior_model: PriorModel, w: cp.Variable) -> cp.Expression:
        """Expression of the portfolio returns series.

        Parameters
        ----------
        prior_model : PriorModel
            The prior model of the assets distributions.

        w : cvxpy Variable
            The CVXPY Variable representing assets weights.

        Returns
        -------
        expression : cvxpy Expression
            The CVXPY Expression the portfolio returns series.
        """
        returns = prior_model.returns @ w
        return returns

    @cache_method("_cvx_cache")
    def _turnover(
        self, n_assets: int, w: cp.Variable, factor: skt.Factor
    ) -> cp.Expression:
        """Expression of the portfolio turnover.

        Parameters
        ----------
        n_assets : int
            The number of assets.

        w : cvxpy Variable
            The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
            Additional variable used for the optimization of some objective function
            like the ratio maximization.

        Returns
        -------
        expression : cvxpy Expression
            The CVXPY Expression the portfolio turnover.
        """
        if self.previous_weights is None:
            raise ValueError(
                "If you provide `max_turnover`, you must also provide "
                " `previous_weights`"
            )
        previous_weights = self._clean_input(
            self.previous_weights,
            n_assets=n_assets,
            fill_value=0,
            name="previous_weights",
        )
        if np.isscalar(previous_weights):
            previous_weights *= np.ones(n_assets)
        turnover = cp.abs(w - previous_weights * factor)
        return turnover

    @cache_method("_cvx_cache")
    def _cvx_min_acceptable_return(
        self,
        prior_model: PriorModel,
        w: cp.Variable,
        min_acceptable_return: skt.Target = None,
    ) -> cp.Expression:
        """Expression of the portfolio Minimum Acceptable Returns.

        Parameters
        ----------
        prior_model : PriorModel
            The prior model of the assets distributions..

        w : cvxpy Variable
            The CVXPY Variable representing assets weights.

        min_acceptable_return : float | ndarray of shape (n_assets,)
            The minimum acceptable return used to distinguish "downside" and "upside"
            returns for the computation of lower partial moments.

        Returns
        -------
        expression : cvxpy Expression
            The CVXPY Expression the portfolio Minimum Acceptable Returns.
        """
        if min_acceptable_return is None:
            min_acceptable_return = prior_model.mu
        if not np.isscalar(min_acceptable_return) and min_acceptable_return.shape != (
            len(min_acceptable_return),
            1,
        ):
            min_acceptable_return = min_acceptable_return[np.newaxis, :]
        mar = (prior_model.returns - min_acceptable_return) @ w
        return mar

    @cache_method("_cvx_cache")
    def __cvx_drawdown(
        self, prior_model: PriorModel, w: cp.Variable, factor: skt.Factor
    ) -> tuple[cp.Variable, list[cp.Expression]]:
        """Expression of the portfolio drawdown.

        Parameters
        ----------
        prior_model : PriorModel
            The prior model of the assets distributions.

        w : cvxpy Variable
            The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
            Additional variable used for the optimization of some objective function
            like the ratio maximization.

        Returns
        -------
        expression : cvxpy Expression
            The CVXPY Expression the portfolio drawdown.
        """
        n_observations = prior_model.returns.shape[0]
        ptf_returns = self._cvx_returns(prior_model=prior_model, w=w)
        ptf_transaction_cost = self._cvx_transaction_cost(
            prior_model=prior_model, w=w, factor=factor
        )
        ptf_management_fee = self._cvx_management_fee(prior_model=prior_model, w=w)
        v = cp.Variable(n_observations + 1)
        constraints = [
            v[1:] * self._scale_constraints
            >= v[:-1] * self._scale_constraints
            - ptf_returns * self._scale_constraints
            + ptf_transaction_cost * self._scale_constraints
            + ptf_management_fee * self._scale_constraints,
            v[1:] * self._scale_constraints >= 0,
            v[0] * self._scale_constraints == 0,
        ]
        return v, constraints

    def _cvx_drawdown(
        self, prior_model: PriorModel, w: cp.Variable, factor: skt.Factor
    ) -> tuple[cp.Variable, list[cp.Expression]]:
        """Expression of the portfolio drawdown.
        Wrapper around __cvx_drawdown to avoid re-adding the constraints when they
        have already been included in the problem.

        Parameters
        ----------
        prior_model : PriorModel
            The prior model of the assets distributions.

        w : cvxpy Variable
            The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
            Additional variable used for the optimization of some objective function
            like the ratio maximization.

        Returns
        -------
        expression : cvxpy Expression
            The CVXPY Expression the portfolio drawdown.
        """
        if "__cvx_drawdown" in self._cvx_cache:
            v, _ = self.__cvx_drawdown(prior_model=prior_model, w=w, factor=factor)
            return v, []
        return self.__cvx_drawdown(prior_model=prior_model, w=w, factor=factor)

    def _tracking_error(
        self, prior_model: PriorModel, w: cp.Variable, y: np.ndarray, factor: skt.Factor
    ) -> cp.Expression:
        """Expression of the portfolio tracking error.

        Parameters
        ----------
        prior_model : PriorModel
            The prior model of the assets distributions.

        w : cvxpy Variable
            The CVXPY Variable representing assets weights.

        y : ndarray of shape (n_observations,)
            Benchmark for the tracking error computation.

        factor : cvxpy Variable | cvxpy Constant
            Additional variable used for the optimization of some objective function
            like the ratio maximization.

        Returns
        -------
        expression : cvxpy Expression
            The CVXPY Expression the portfolio tracking error.
        """
        n_observations = prior_model.returns.shape[0]
        ptf_returns = self._cvx_returns(prior_model=prior_model, w=w)
        tracking_error = cp.norm(ptf_returns - y * factor, "fro") / cp.sqrt(
            n_observations - 1
        )
        return tracking_error

    # Risk Measures risk models
    # They need to be named f'_{risk_measure}_risk' as they are loaded dynamically in
    # mean_risk_optimization()
    def _mean_absolute_deviation_risk(
        self, prior_model: PriorModel, w: cp.Variable, min_acceptable_return: skt.Target
    ) -> skt.RiskResult:
        """Expression and Constraints of the Mean Absolute Deviation risk measure.

        Parameters
        ----------
        prior_model : PriorModel
            The prior model of the assets distributions.

        w : cvxpy Variable
            The CVXPY Variable representing assets weights.

        min_acceptable_return : float | ndarray of shape (n_assets,)
            The minimum acceptable return used to distinguish "downside" and "upside"
            returns for the computation of lower partial moments.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
            CVXPY Expression and Constraints of the Mean Absolute Deviation risk
            measure.
        """
        n_observations = prior_model.returns.shape[0]
        ptf_min_acceptable_return = self._cvx_min_acceptable_return(
            prior_model=prior_model, w=w, min_acceptable_return=min_acceptable_return
        )
        v = cp.Variable(n_observations, nonneg=True)
        risk = 2 * cp.sum(v) / n_observations
        constraints = [
            ptf_min_acceptable_return * self._scale_constraints
            >= -v * self._scale_constraints
        ]
        return risk, constraints

    def _first_lower_partial_moment_risk(
        self,
        prior_model: PriorModel,
        w: cp.Variable,
        min_acceptable_return: skt.Target,
        factor: skt.Factor,
    ) -> skt.RiskResult:
        """Expression and Constraints of the First Lower Partial Moment risk measure.

        Parameters
        ----------
        prior_model : PriorModel
            The prior model of the assets distributions.

        w : cvxpy Variable
            The CVXPY Variable representing assets weights.

        min_acceptable_return : float | ndarray of shape (n_assets,)
            The minimum acceptable return used to distinguish "downside" and "upside"
            returns for the computation of lower partial moments.

        factor : cvxpy Variable | cvxpy Constant
            Additional variable used for the optimization of some objective function
            like the ratio maximization.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
            CVXPY Expression and Constraints of the First Lower Partial Moment risk
            measure.
        """
        n_observations = prior_model.returns.shape[0]
        ptf_min_acceptable_return = self._cvx_min_acceptable_return(
            prior_model=prior_model, w=w, min_acceptable_return=min_acceptable_return
        )
        v = cp.Variable(n_observations, nonneg=True)
        risk = cp.sum(v) / n_observations
        constraints = [
            self.risk_free_rate * factor * self._scale_constraints
            - ptf_min_acceptable_return * self._scale_constraints
            <= v * self._scale_constraints
        ]
        return risk, constraints

    def _standard_deviation_risk(
        self, prior_model: PriorModel, w: cp.Variable
    ) -> skt.RiskResult:
        """Expression and Constraints of the Standard Deviation risk measure.

        Parameters
        ----------
        prior_model : PriorModel
            The prior model of the assets distributions.

        w : cvxpy Variable
            The CVXPY Variable representing assets weights.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
            CVXPY Expression and Constraints of the Standard Deviation risk measure.
        """
        v = cp.Variable(
            nonneg=True
        )  # nonneg=True instead of constraint v>=0 is preferred for better DCP analysis
        if prior_model.cholesky is not None:
            z = prior_model.cholesky
        else:
            z = np.linalg.cholesky(prior_model.covariance)
        risk = v
        constraints = [
            cp.SOC(v * self._scale_constraints, z.T @ w * self._scale_constraints)
        ]
        return risk, constraints

    def _variance_risk(self, prior_model: PriorModel, w: cp.Variable) -> skt.RiskResult:
        """Expression and Constraints of the Variance risk measure.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
           CVXPY Expression and Constraints the Variance risk measure.
        """
        risk, constraints = self._standard_deviation_risk(prior_model=prior_model, w=w)
        risk = cp.square(risk)
        return risk, constraints

    def _worst_case_variance_risk(
        self,
        prior_model: PriorModel,
        covariance_uncertainty_set: UncertaintySet,
        w: cp.Variable,
        factor: skt.Factor,
    ) -> skt.RiskResult:
        """Expression and Constraints of the Worst Case Variance.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        covariance_uncertainty_set : UncertaintySet
             :ref:`Covariance Uncertainty set estimator <uncertainty_set_estimator>`.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
           Additional variable used for the optimization of some objective function
           like the ratio maximization.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
           CVXPY Expression and Constraints the Worst Case Variance.
        """
        n_assets = prior_model.returns.shape[1]
        x = cp.Variable((n_assets, n_assets), symmetric=True)
        y = cp.Variable((n_assets, n_assets), symmetric=True)
        w_reshaped = cp.reshape(w, (n_assets, 1))
        factor_reshaped = cp.reshape(factor, (1, 1))
        z1 = cp.vstack([x, w_reshaped.T])
        z2 = cp.vstack([w_reshaped, factor_reshaped])

        risk = covariance_uncertainty_set.k * cp.pnorm(
            sc.linalg.sqrtm(covariance_uncertainty_set.sigma) @ (cp.vec(x) + cp.vec(y)),
            2,
        ) + cp.trace(prior_model.covariance @ (x + y))
        # semi-definite positive constraints
        # noinspection PyTypeChecker
        constraints = [
            cp.hstack([z1, z2]) * self._scale_constraints >> 0,
            y * self._scale_constraints >> 0,
        ]
        return risk, constraints

    def _semi_variance_risk(
        self,
        prior_model: PriorModel,
        w: cp.Variable,
        min_acceptable_return: skt.Target = None,
    ) -> skt.RiskResult:
        """Expression and Constraints of the Semi Variance risk measure.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        min_acceptable_return : float | ndarray of shape (n_assets,)
            The minimum acceptable return used to distinguish "downside" and "upside"
            returns for the computation of lower partial moments.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
            CVXPY Expression and Constraints the Semi Variance risk measure.
        """
        n_observations = prior_model.returns.shape[0]
        ptf_min_acceptable_return = self._cvx_min_acceptable_return(
            prior_model=prior_model, w=w, min_acceptable_return=min_acceptable_return
        )
        v = cp.Variable(n_observations, nonneg=True)
        risk = cp.sum_squares(v) / (n_observations - 1)
        constraints = [
            ptf_min_acceptable_return * self._scale_constraints
            >= -v * self._scale_constraints
        ]
        return risk, constraints

    def _semi_deviation_risk(
        self,
        prior_model: PriorModel,
        w: cp.Variable,
        min_acceptable_return: skt.Target = None,
    ) -> skt.RiskResult:
        """Expression and Constraints of the Semi Standard Deviation risk measure.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        min_acceptable_return : float | ndarray of shape (n_assets,)
            The minimum acceptable return used to distinguish "downside" and "upside"
            returns for the computation of lower partial moments.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
            CVXPY Expression and Constraints the Semi Standard Deviation risk measure.
        """
        n_observations = prior_model.returns.shape[0]
        ptf_min_acceptable_return = self._cvx_min_acceptable_return(
            prior_model=prior_model, w=w, min_acceptable_return=min_acceptable_return
        )
        v = cp.Variable(n_observations, nonneg=True)
        risk = cp.norm(v, 2) / np.sqrt(n_observations - 1)
        constraints = [
            ptf_min_acceptable_return * self._scale_constraints
            >= -v * self._scale_constraints
        ]
        return risk, constraints

    def _fourth_central_moment_risk(self, w: cp.Variable, factor: skt.Factor):
        raise NotImplementedError

    def _fourth_lower_partial_moment_risk(self, w: cp.Variable, factor: skt.Factor):
        raise NotImplementedError

    def _worst_realization_risk(
        self, prior_model: PriorModel, w: cp.Variable, factor: skt.Factor
    ) -> skt.RiskResult:
        """Expression and Constraints of the Worst Realization risk measure.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
           Additional variable used for the optimization of some objective function
           like the ratio maximization.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
            CVXPY Expression and Constraints the Worst Realization risk measure.
        """
        ptf_returns = self._cvx_returns(prior_model=prior_model, w=w)
        ptf_transaction_cost = self._cvx_transaction_cost(
            prior_model=prior_model, w=w, factor=factor
        )
        ptf_management_fee = self._cvx_management_fee(prior_model=prior_model, w=w)
        v = cp.Variable()
        risk = v
        constraints = [
            -ptf_returns * self._scale_constraints
            + ptf_transaction_cost * self._scale_constraints
            + ptf_management_fee * self._scale_constraints
            <= v * self._scale_constraints
        ]
        return risk, constraints

    def _cvar_risk(
        self,
        prior_model: PriorModel,
        w: cp.Variable,
        factor: skt.Factor,
    ) -> skt.RiskResult:
        """Expression and Constraints of the CVaR risk measure.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
           Additional variable used for the optimization of some objective function
           like the ratio maximization.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
            CVXPY Expression and Constraints the CVaR risk measure.
        """
        n_observations = prior_model.returns.shape[0]
        ptf_returns = self._cvx_returns(prior_model=prior_model, w=w)
        ptf_transaction_cost = self._cvx_transaction_cost(
            prior_model=prior_model, w=w, factor=factor
        )
        ptf_management_fee = self._cvx_management_fee(prior_model=prior_model, w=w)
        alpha = cp.Variable()
        v = cp.Variable(n_observations, nonneg=True)
        risk = alpha + 1.0 / (n_observations * (1 - self.cvar_beta)) * cp.sum(v)
        # noinspection PyTypeChecker
        constraints = [
            ptf_returns * self._scale_constraints
            - ptf_transaction_cost * self._scale_constraints
            - ptf_management_fee * self._scale_constraints
            + alpha * self._scale_constraints
            + v * self._scale_constraints
            >= 0
        ]
        return risk, constraints

    def _evar_risk(
        self,
        prior_model: PriorModel,
        w: cp.Variable,
        factor: skt.Factor,
    ) -> skt.RiskResult:
        """Expression and Constraints of the EVaR risk measure.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
           Additional variable used for the optimization of some objective function
           like the ratio maximization.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
            CVXPY Expression and Constraints the EVaR risk measure.
        """
        n_observations = prior_model.returns.shape[0]
        ptf_returns = self._cvx_returns(prior_model=prior_model, w=w)
        ptf_transaction_cost = self._cvx_transaction_cost(
            prior_model=prior_model, w=w, factor=factor
        )
        ptf_management_fee = self._cvx_management_fee(prior_model=prior_model, w=w)
        # We don't include the transaction_cost in the constraint otherwise the problem
        # is not DCP
        if not isinstance(ptf_transaction_cost, cp.Constant):
            warnings.warn(
                "The EVaR problem will be relaxed by removing the transaction costs"
                " from the Cone constraint to keep the problem DCP. The solution may"
                " not be accurate.",
                stacklevel=2,
            )

        x = cp.Variable()
        y = cp.Variable(nonneg=True)
        z = cp.Variable(n_observations)
        risk = x + y * np.log(1 / (n_observations * (1 - self.evar_beta)))
        constraints = [
            cp.sum(z) * self._scale_constraints <= y * self._scale_constraints,
            cp.constraints.ExpCone(
                -ptf_returns * self._scale_constraints
                + ptf_management_fee * self._scale_constraints
                - x * self._scale_constraints,
                np.ones(n_observations) * y * self._scale_constraints,
                z * self._scale_constraints,
            ),
        ]
        return risk, constraints

    def _max_drawdown_risk(
        self, prior_model: PriorModel, w: cp.Variable, factor: skt.Factor
    ) -> skt.RiskResult:
        """Expression and Constraints of the EVaR risk measure.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
           Additional variable used for the optimization of some objective function
           like the ratio maximization.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
            CVXPY Expression and Constraints the EVaR risk measure.
        """
        v, constraints = self._cvx_drawdown(prior_model=prior_model, w=w, factor=factor)
        u = cp.Variable()
        risk = u
        constraints += [u * self._scale_constraints >= v[1:] * self._scale_constraints]
        return risk, constraints

    def _average_drawdown_risk(
        self, prior_model: PriorModel, w: cp.Variable, factor: skt.Factor
    ) -> skt.RiskResult:
        """Expression and Constraints of the Average Drawdown risk measure.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
           Additional variable used for the optimization of some objective function
           like the ratio maximization.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
            CVXPY Expression and Constraints the Average Drawdown risk measure.
        """
        n_observations = prior_model.returns.shape[0]
        v, constraints = self._cvx_drawdown(prior_model=prior_model, w=w, factor=factor)
        risk = cp.sum(v[1:]) / n_observations
        return risk, constraints

    def _cdar_risk(
        self,
        prior_model: PriorModel,
        w: cp.Variable,
        factor: skt.Factor,
    ) -> skt.RiskResult:
        """Expression and Constraints of the CDaR risk measure.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
           Additional variable used for the optimization of some objective function
           like the ratio maximization.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
            CVXPY Expression and Constraints the CDaR risk measure.
        """
        n_observations = prior_model.returns.shape[0]
        v, constraints = self._cvx_drawdown(prior_model=prior_model, w=w, factor=factor)
        alpha = cp.Variable()
        z = cp.Variable(n_observations, nonneg=True)
        risk = alpha + 1.0 / (n_observations * (1 - self.cdar_beta)) * cp.sum(z)
        constraints += [
            z * self._scale_constraints
            >= v[1:] * self._scale_constraints - alpha * self._scale_constraints
        ]
        return risk, constraints

    def _edar_risk(
        self,
        prior_model: PriorModel,
        w: cp.Variable,
        factor: skt.Factor,
    ) -> skt.RiskResult:
        """Expression and Constraints of the EDaR risk measure.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
           Additional variable used for the optimization of some objective function
           like the ratio maximization.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
            CVXPY Expression and Constraints the EDaR risk measure.
        """
        n_observations = prior_model.returns.shape[0]
        v, constraints = self._cvx_drawdown(prior_model=prior_model, w=w, factor=factor)
        x = cp.Variable()
        y = cp.Variable(nonneg=True)
        z = cp.Variable(n_observations)
        risk = x + y * np.log(1 / (n_observations * (1 - self.edar_beta)))
        constraints += [
            cp.sum(z) * self._scale_constraints <= y * self._scale_constraints,
            cp.constraints.ExpCone(
                v[1:] * self._scale_constraints - x * self._scale_constraints,
                np.ones(n_observations) * y * self._scale_constraints,
                z * self._scale_constraints,
            ),
        ]
        return risk, constraints

    def _ulcer_index_risk(
        self,
        prior_model: PriorModel,
        w: cp.Variable,
        factor: skt.Factor,
    ) -> skt.RiskResult:
        """Expression and Constraints of the Ulcer Index risk measure.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
           Additional variable used for the optimization of some objective function
           like the ratio maximization.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
            CVXPY Expression and Constraints the Ulcer Index risk measure.
        """
        v, constraints = self._cvx_drawdown(prior_model=prior_model, w=w, factor=factor)
        n_observations = prior_model.returns.shape[0]
        risk = cp.norm(v[1:], 2) / (np.sqrt(n_observations))
        return risk, constraints

    def _gini_mean_difference_risk(
        self, prior_model: PriorModel, w: cp.Variable, factor: skt.Factor
    ) -> skt.RiskResult:
        """Expression and Constraints of the Gini Mean Difference risk measure.

        The Gini mean difference (GMD) is a measure of dispersion introduced in the
        context of portfolio optimization by Yitzhaki (1982).
        The initial formulation was not used by practitioners due to the high number of
        variables that increases proportional to T(T-1)/2 ,

        Cajas (2021) proposed an alternative reformulation based on the ordered weighted
        averaging (OWA) operator for monotonic weights proposed by Chassein and
        Goerigk (2015). We implement this formulation which is more efficient for large
        scale problems.

        Parameters
        ----------
        prior_model : PriorModel
           The prior model of the assets distributions.

        w : cvxpy Variable
           The CVXPY Variable representing assets weights.

        factor : cvxpy Variable | cvxpy Constant
           Additional variable used for the optimization of some objective function
           like the ratio maximization.

        Returns
        -------
        expression : tuple[cvxpy Expression , list[cvxpy Expression]]
            CVXPY Expression and Constraints the Ulcer Index risk measure.
        """
        ptf_returns = self._cvx_returns(prior_model=prior_model, w=w)
        ptf_transaction_cost = self._cvx_transaction_cost(
            prior_model=prior_model, w=w, factor=factor
        )
        ptf_management_fee = self._cvx_management_fee(prior_model=prior_model, w=w)
        observation_nb = prior_model.returns.shape[0]
        x = cp.Variable((observation_nb, 1))
        y = cp.Variable((observation_nb, 1))
        z = cp.Variable((observation_nb, 1))
        ones = np.ones((observation_nb, 1))
        risk = 2 * cp.sum(x + y)
        gmd_w = np.array(owa_gmd_weights(observation_nb) / 2).reshape(-1, 1)
        # noinspection PyTypeChecker
        constraints = [
            ptf_returns * self._scale_constraints
            - ptf_transaction_cost * self._scale_constraints
            - ptf_management_fee * self._scale_constraints
            == cp.reshape(z, (observation_nb,)) * self._scale_constraints,
            z @ gmd_w.T <= ones @ x.T + y @ ones.T,
        ]
        return risk, constraints

    @abstractmethod
    def fit(self, X: npt.ArrayLike, y: npt.ArrayLike | None = None):
        pass



"""Factor Model estimator"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# Riskfolio-Lib, Copyright (c) 2020-2023, Dany Cajas, Licensed under BSD 3 clause.
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

from abc import ABC, abstractmethod
from typing import Any

import numpy as np
import numpy.typing as npt
import sklearn.base as skb
import sklearn.linear_model as skl
import sklearn.multioutput as skm

from skfolio.prior._base import BasePrior, PriorModel
from skfolio.prior._empirical import EmpiricalPrior
from skfolio.utils.stats import cov_nearest
from skfolio.utils.tools import check_estimator


class BaseLoadingMatrix(skb.BaseEstimator, ABC):
    """Base class for all Loading Matrix estimators.

    Notes
    -----
    All estimators should specify all the parameters that can be set
    at the class level in their ``__init__`` as explicit keyword
    arguments (no ``*args`` or ``**kwargs``).
    """

    loading_matrix_: np.ndarray
    intercepts_: np.ndarray

    @abstractmethod
    def fit(self, X: npt.ArrayLike, y: npt.ArrayLike):
        pass


class LoadingMatrixRegression(BaseLoadingMatrix):
    """Loading Matrix Regression estimator.

    Estimate the loading matrix by fitting one linear regressor per asset.

    Parameters
    ----------
    linear_regressor : BaseEstimator, optional
       Linear regressor used to fit the factors on each asset separately.
       The default (`None`) is to use `LassoCV(fit_intercept=False)`.

    n_jobs : int, optional
        The number of jobs to run in parallel.

        When individual estimators are fast to train or predict,
        using ``n_jobs > 1`` can result in slower performance due
        to the parallelism overhead.

        The value `-1` means using all processors.
        The default (`None`) means 1 unless in a `joblib.parallel_backend` context.

    Attributes
    ----------
    loading_matrix_ : ndarray of shape (n_assets, n_factors)
        The loading matrix.

    intercepts_: ndarray of shape (n_assets,)
        The intercepts.

    multi_output_regressor_: MultiOutputRegressor
        Fitted `sklearn.multioutput.MultiOutputRegressor`
    """

    multi_output_regressor_: skm.MultiOutputRegressor

    def __init__(
        self,
        linear_regressor: skb.BaseEstimator | None = None,
        n_jobs: int | None = None,
    ):
        self.linear_regressor = linear_regressor
        self.n_jobs = n_jobs

    def fit(self, X: npt.ArrayLike, y: npt.ArrayLike):
        """Fit the Loading Matrix Regression Estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : array-like of shape (n_observations, n_factors)
            Price returns of the factors.

        Returns
        -------
        self : LoadingMatrixRegression
            Fitted estimator.
        """
        _linear_regressor = check_estimator(
            self.linear_regressor,
            default=skl.LassoCV(fit_intercept=False),
            check_type=skb.BaseEstimator,
        )

        self.multi_output_regressor_ = skm.MultiOutputRegressor(
            _linear_regressor, n_jobs=self.n_jobs
        )
        self.multi_output_regressor_.fit(X=y, y=X)
        # noinspection PyUnresolvedReferences
        n_assets = X.shape[1]
        self.loading_matrix_ = np.array(
            [self.multi_output_regressor_.estimators_[i].coef_ for i in range(n_assets)]
        )
        self.intercepts_ = np.array(
            [
                self.multi_output_regressor_.estimators_[i].intercept_
                for i in range(n_assets)
            ]
        )


class FactorModel(BasePrior):
    """Factor Model estimator.

    The purpose of Factor Models is to impose a structure on financial variables and
    their covariance matrix by explaining them through a small number of common factors.
    This can help overcome estimation error by reducing the number of parameters,
    i.e. the dimensionality of the estimation problem, making portfolio optimization
    more robust against noise in the data. Factor Models also provide a decomposition of
    financial risk into systematic and security-specific components.

    Parameters
    ----------
    loading_matrix_estimator : LoadingMatrixEstimator, optional
        Estimator of the loading matrix (betas) of the factors.
        The default (`None`) is to use :class:`LoadingMatrixRegression` which fit the
        factors using `LassoCV` on each asset separately.

    factor_prior_estimator : BasePrior, optional
        The factors :ref:`prior estimator <prior>`.
        It is used to estimate the :class:`~skfolio.prior.PriorModel` containing the
        factors expected returns and covariance matrix.
        The default (`None`) is to use :class:`~skfolio.prior.EmpiricalPrior`.

    residual_variance : bool, default=True
        If this is set to True, the diagonal term of the residuals covariance
        (residuals variance) is added to the factor model covariance.

    higham : bool, default=False
        If this is set to True, we use the Higham & Nick (2002) algorithm to find the
        nearest covariance matrix that is positive semi-definite. It is more accurate
        but slower that the default clipping method. For more information
        see :func:`~skfolio.utils.stats.cov_nearest`.

    max_iteration : int, default=100
        Only used when `higham` is set to True. Maximum number of iterations of the
        Higham & Nick (2002) algorithm.

    Attributes
    ----------
    prior_model_ : PriorModel
        The :class:`~skfolio.prior.PriorModel`.

    factor_prior_estimator_ : BasePrior
        Fitted `factor_prior_estimator`.

    loading_matrix_estimator_ : BaseLoadingMatrix
        Fitted `loading_matrix_estimator`.

    n_features_in_ : int
       Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
       Names of features seen during `fit`. Defined only when `X`
       has feature names that are all strings.
    """

    factor_prior_estimator_: BasePrior
    loading_matrix_estimator_: BaseLoadingMatrix

    def __init__(
        self,
        loading_matrix_estimator: BaseLoadingMatrix | None = None,
        factor_prior_estimator: BasePrior | None = None,
        residual_variance: bool = True,
        higham: bool = False,
        max_iteration: int = 100,
    ):
        self.loading_matrix_estimator = loading_matrix_estimator
        self.factor_prior_estimator = factor_prior_estimator
        self.residual_variance = residual_variance
        self.higham = higham
        self.max_iteration = max_iteration

    # noinspection PyMethodOverriding, PyPep8Naming
    def fit(self, X: npt.ArrayLike, y: Any):
        """Fit the Factor Model estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : array-like of shape (n_observations, n_factors)
            Factors' returns.

        Returns
        -------
        self : FactorModel
            Fitted estimator.
        """
        self.factor_prior_estimator_ = check_estimator(
            self.factor_prior_estimator,
            default=EmpiricalPrior(),
            check_type=BasePrior,
        )
        self.loading_matrix_estimator_ = check_estimator(
            self.loading_matrix_estimator,
            default=LoadingMatrixRegression(),
            check_type=BaseLoadingMatrix,
        )

        # Fitting prior estimator
        self.factor_prior_estimator_.fit(y)
        factor_mu = self.factor_prior_estimator_.prior_model_.mu
        factor_covariance = self.factor_prior_estimator_.prior_model_.covariance

        # Fitting loading matrix estimator
        self.loading_matrix_estimator_.fit(X, y)
        loading_matrix = self.loading_matrix_estimator_.loading_matrix_
        intercepts = self.loading_matrix_estimator_.intercepts_

        # we validate and convert to numpy after all models have been fitted to keep
        # features names information.
        X, y = self._validate_data(X, y, multi_output=True)
        n_assets = X.shape[1]
        n_factors = y.shape[1]

        if loading_matrix.shape != (n_assets, n_factors):
            raise ValueError(
                "`loading_matrix_estimator.loading_matrix_` must ba a 2D array of"
                f" shape {(n_assets, n_factors)}, got"
                f" {loading_matrix.shape} instead."
            )

        if intercepts.shape != (n_assets,):
            raise ValueError(
                "`loading_matrix_estimator.intercepts_` must ba a 1D array of "
                f"shape {(n_assets,)}, got {intercepts.shape} instead."
            )

        mu = loading_matrix @ factor_mu + intercepts
        covariance = loading_matrix @ factor_covariance @ loading_matrix.T
        returns = y @ loading_matrix.T + intercepts
        cholesky = loading_matrix @ np.linalg.cholesky(factor_covariance)

        if self.residual_variance:
            err = X - returns
            err_cov = np.diag(np.var(err, ddof=1, axis=0))
            covariance += err_cov
            cholesky = np.hstack((cholesky, np.sqrt(err_cov)))

        covariance = cov_nearest(
            covariance, higham=self.higham, higham_max_iteration=self.max_iteration
        )

        self.prior_model_ = PriorModel(
            mu=mu, covariance=covariance, returns=returns, cholesky=cholesky
        )
        return self



"""Combinatorial module"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# scikit-portfolio, Copyright (c) 2022, Carlo Nicolini, Licensed under MIT Licence.
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

import itertools
import math
import numbers
from abc import ABC, abstractmethod
from collections.abc import Iterator

import numpy as np
import numpy.typing as npt
import pandas as pd
import plotly.graph_objects as go
import sklearn.model_selection as skm
import sklearn.utils as sku

import skfolio.typing as skt


class BaseCombinatorialCV(ABC):
    """Base class for all combinatorial cross-validators.

    Implementations must define `split` or `get_path_ids`.
    """

    @abstractmethod
    def split(self, X: npt.ArrayLike, y=None) -> tuple[np.ndarray, list[np.ndarray]]:
        pass

    @abstractmethod
    def get_path_ids(self) -> np.ndarray:
        """Return the path id of each test sets in each split"""
        pass

    __repr__ = skm.BaseCrossValidator.__repr__


# TODO: review params and function naming
class CombinatorialPurgedCV(BaseCombinatorialCV):
    """Combinatorial Purged Cross-Validation.

    Provides train/test indices to split time series data samples based on
    Combinatorial Purged Cross-Validation [1]_.

    Compared to `KFold`, which splits the data into `k` folds with `1` fold for the test
    set and `k - 1` folds for the training set, `CombinatorialPurgedCV` uses `k - p`
    folds for the training set with `p > 1` being the number of test folds.

    `KFold` can recombine one single testing path while `CombinatorialPurgedCV` can
    recombine multiple testing paths from the combinations of the train/test sets.

    To avoid data leakage, purging and embargoing can be performed.

    Purging consist of removing from the training set all observations whose labels
    overlapped in time with those labels included in the testing set.

    Embargoing consist of removing from the training set all observations that
    immediately follow an observation in the testing set, since financial features
    often incorporate series that exhibit serial correlation (like ARMA processes).

    Parameters
    ----------
    n_folds : int, default=10
        Number of folds. Must be at least 3.

    n_test_folds : int, default=8
        Number of test folds. Must be at least 2.
        For only one test fold, use `sklearn.model_validation.KFold`.

    purged_size : int, default=0
        Number of observations to exclude from the start of each train set that are
        after a test set **and** the number of observations to exclude from the end of
        each training set that are before a test set.

    embargo_size : int, default=0
        Number of observations to exclude from the start of each training set that are
        after a test set.

    Attributes
    ----------
    index_train_test_ : ndarray of shape (n_observations, n_splits)

    Examples
    --------
    >>> import numpy as np
    >>> from skfolio.model_selection import CombinatorialPurgedCV
    >>> X = np.random.randn(12, 2)
    >>> cv = CombinatorialPurgedCV(n_folds=3, n_test_folds=2)
    >>> for i, (train_index, tests) in enumerate(cv.split(X)):
    ...     print(f"Split {i}:")
    ...     print(f"  Train: index={train_index}")
    ...     for j, test_index in enumerate(tests):
    ...         print(f"  Test {j}:  index={test_index}")
    Split 0:
      Train: index=[ 8  9 10 11]
      Test 0:  index=[0 1 2 3]
      Test 1:  index=[4 5 6 7]
    Split 1:
      Train: index=[4 5 6 7]
      Test 0:  index=[0 1 2 3]
      Test 1:  index=[ 8  9 10 11]
    Split 2:
      Train: index=[0 1 2 3]
      Test 0:  index=[4 5 6 7]
      Test 1:  index=[ 8  9 10 11]
    >>> cv = CombinatorialPurgedCV(n_folds=3, n_test_folds=2, purged_size=1)
    >>> for i, (train_index, tests) in enumerate(cv.split(X)):
    ...     print(f"Split {i}:")
    ...     print(f"  Train: index={train_index}")
    ...     for j, test_index in enumerate(tests):
    ...         print(f"  Test {j}:  index={test_index}")
    Split 0:
      Train: index=[ 9 10 11]
      Test 0:  index=[0 1 2 3]
      Test 1:  index=[4 5 6 7]
    Split 1:
      Train: index=[5 6]
      Test 0:  index=[0 1 2 3]
      Test 1:  index=[ 8  9 10 11]
    Split 2:
      Train: index=[0 1 2]
      Test 0:  index=[4 5 6 7]
      Test 1:  index=[ 8  9 10 11]
    >>> cv = CombinatorialPurgedCV(n_folds=3, n_test_folds=2, embargo_size=1)
    >>> for i, (train_index, tests) in enumerate(cv.split(X)):
    ...     print(f"Split {i}:")
    ...     print(f"  Train: index={train_index}")
    ...     for j, test_index in enumerate(tests):
    ...         print(f"  Test {j}:  index={test_index}")
    Split 0:
      Train: index=[ 9 10 11]
      Test 0:  index=[0 1 2 3]
      Test 1:  index=[4 5 6 7]
    Split 1:
      Train: index=[5 6 7]
      Test 0:  index=[0 1 2 3]
      Test 1:  index=[ 8  9 10 11]
    Split 2:
      Train: index=[0 1 2 3]
      Test 0:  index=[4 5 6 7]
      Test 1:  index=[ 8  9 10 11]

    References
    ----------
    .. [1]  "Advances in Financial Machine Learning",
        Marcos López de Prado (2018)
    """

    index_train_test_: np.ndarray

    def __init__(
        self,
        n_folds: int = 10,
        n_test_folds: int = 8,
        purged_size: int = 0,
        embargo_size: int = 0,
    ):
        if not isinstance(n_folds, numbers.Integral):
            raise ValueError(
                "The number of folds must be of Integral type. "
                f"{n_folds} of type {type(n_folds)} was passed."
            )
        n_folds = int(n_folds)

        if n_folds <= 2:
            raise ValueError(f"`n_folds` must be at least 3`, got `n_folds={n_folds}`.")

        if n_test_folds <= 1:
            raise ValueError(
                f"`n_test_folds` must at least 2, got `n_test_folds={n_test_folds}`."
            )

        if n_test_folds >= n_folds:
            raise ValueError(
                "Combinatorial purged cross-validation requires `n_folds` "
                "to be greater than `n_test_folds`."
            )

        if purged_size < 0:
            raise ValueError("`purged_size` cannot be negative")

        if embargo_size < 0:
            raise ValueError("`embargo_size` cannot be negative")

        self.n_folds = n_folds
        self.n_test_folds = n_test_folds
        self.purged_size = purged_size
        self.embargo_size = embargo_size

    @property
    def n_splits(self) -> int:
        """Number of splits"""
        return _n_splits(n_folds=self.n_folds, n_test_folds=self.n_test_folds)

    @property
    def n_test_paths(self) -> int:
        """Number of test paths that can be reconstructed from the train/test
        combinations"""
        return _n_test_paths(n_folds=self.n_folds, n_test_folds=self.n_test_folds)

    @property
    def test_set_index(self) -> np.ndarray:
        """Location of each test set"""
        return np.array(
            list(itertools.combinations(np.arange(self.n_folds), self.n_test_folds))
        ).reshape(-1, self.n_test_folds)

    @property
    def binary_train_test_sets(self) -> np.ndarray:
        """Identify training and test folds for each combinations by assigning `0` to
        training folds and `1` to test folds"""
        folds_train_test = np.zeros((self.n_folds, self.n_splits))
        folds_train_test[
            self.test_set_index, np.arange(self.n_splits)[:, np.newaxis]
        ] = 1
        return folds_train_test

    @property
    def recombined_paths(self) -> np.ndarray:
        """Recombine each test path by returning the test set location in each split."""
        return np.argwhere(self.binary_train_test_sets == 1)[:, 1].reshape(
            self.n_folds, -1
        )

    def get_path_ids(self) -> np.ndarray:
        """Return the path id of each test sets in each split"""
        recombine_paths = self.recombined_paths
        path_ids = np.zeros((self.n_splits, self.n_test_folds), dtype=int)
        for i in range(self.n_splits):
            for j in range(self.n_test_folds):
                path_ids[i, j] = np.argwhere(recombine_paths == i)[j][1]
        return path_ids

    def split(
        self, X: npt.ArrayLike, y=None, groups=None
    ) -> Iterator[tuple[np.ndarray, list[np.ndarray]]]:
        """Generate indices to split data into training and test set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples
            and `n_features` is the number of features.

        y : array-like of shape (n_samples,), optional
            The (multi-)target variable

        groups : array-like of shape (n_samples,), optional
            Group labels for the samples used while splitting the dataset into
            train/test set.

        Yields
        ------
        train : ndarray
            The training set indices for that split.

        test : ndarray
            The testing set indices for that split.
        """
        test_set_index = self.test_set_index
        recombine_paths = self.recombined_paths

        X, y = sku.indexable(X, y)
        n_samples = X.shape[0]
        min_fold_size = n_samples // self.n_folds
        if self.purged_size + self.embargo_size >= min_fold_size - 1:
            raise ValueError(
                "The sum of `purged_size` and `embargo_size` must be smaller than the"
                f" size of a train fold which is {min_fold_size}"
            )

        fold_index_num = np.arange(n_samples) // (n_samples // self.n_folds)
        fold_index_num[fold_index_num == self.n_folds] = self.n_folds - 1

        index_train_test = np.zeros((n_samples, self.n_splits))
        for i in range(self.n_splits):
            index_train_test[
                np.argwhere([fold_index_num == j for j in test_set_index[i]])[:, 1], i
            ] = 1

        diff = np.diff(index_train_test, axis=0)

        # Purge before
        before_index = np.argwhere(diff == 1)
        for k in range(self.purged_size):
            index_train_test[
                np.maximum(0, before_index[:, 0] - k), before_index[:, 1]
            ] = -1

        # Purge after and Embargo
        after_index = np.argwhere(diff == -1)
        for k in range(self.purged_size + self.embargo_size):
            index_train_test[
                np.minimum(n_samples - 1, after_index[:, 0] + k + 1), after_index[:, 1]
            ] = -1
        self.index_train_test_ = index_train_test

        fold_index = {
            fold_id: np.argwhere(fold_index_num == fold_id).reshape(-1)
            for fold_id in range(self.n_folds)
        }
        for i in range(self.n_splits):
            train_index = np.argwhere(index_train_test[:, i] == 0).reshape(-1)
            test_index_list = [
                fold_index[fold_id] for fold_id, _ in np.argwhere(recombine_paths == i)
            ]
            yield train_index, test_index_list

    def summary(self, X) -> pd.Series:
        n_observations = X.shape[0]
        avg_train_size = _avg_train_size(
            n_observations=n_observations,
            n_folds=self.n_folds,
            n_test_folds=self.n_test_folds,
        )
        return pd.Series(
            {
                "Number of Observations": n_observations,
                "Total Number of Folds": self.n_folds,
                "Number of Test Folds": self.n_test_folds,
                "Purge Size": self.purged_size,
                "Embargo Size": self.embargo_size,
                "Average Training Size": int(avg_train_size),
                "Number of Test Paths": self.n_test_paths,
                "Number of Training Combinations": self.n_splits,
            }
        )

    def plot_train_test_folds(self) -> skt.Figure:
        """Plot the train/test fold locations"""
        values = self.binary_train_test_sets
        fill_color = np.where(values == 0, "blue", "red")
        fill_color = fill_color.astype(object)
        fill_color = np.insert(
            fill_color, 0, np.array(["darkblue" for _ in range(self.n_splits)]), axis=0
        )
        values = np.insert(values, 0, np.arange(self.n_splits), axis=0)
        fig = go.Figure(
            data=[
                go.Table(
                    header=dict(
                        values=["Train Combinations"]
                        + [f"Fold {i}" for i in range(self.n_folds)],
                        fill_color="darkblue",
                        font=dict(color="white"),
                        align="left",
                    ),
                    cells=dict(
                        values=values,
                        font=dict(color="white"),
                        fill_color=fill_color,
                        line_color="grey",
                        align="left",
                    ),
                )
            ]
        )
        fig.update_layout(title="Split Train (0) /Test (1) Folds per Combination")
        return fig

    def plot_train_test_index(self, X) -> skt.Figure:
        """Plot the training and test indices for each combinations by assigning `0` to
        training, `1` to test and `-1` to both purge and embargo indices."""
        next(self.split(X))
        n_samples = X.shape[0]
        cond = [
            self.index_train_test_ == -1,
            self.index_train_test_ == 0,
            self.index_train_test_ == 1,
        ]
        values = self.index_train_test_.T
        values = np.insert(values, 0, np.arange(n_samples), axis=0)
        fill_color = np.select(cond, ["green", "blue", "red"]).T
        fill_color = fill_color.astype(object)
        fill_color = np.insert(
            fill_color, 0, np.array(["darkblue" for _ in range(n_samples)]), axis=0
        )
        fig = go.Figure(
            data=[
                go.Table(
                    header=dict(
                        values=["observations"]
                        + [f"Split {i}" for i in range(self.n_splits)],
                        fill_color="darkblue",
                        font=dict(color="white"),
                        align="left",
                    ),
                    cells=dict(
                        values=values,
                        font=dict(color="white"),
                        fill_color=fill_color,
                        line_color="grey",
                        align="left",
                    ),
                )
            ]
        )
        fig.update_layout(
            title="Train (0), Test (1) and Purge/Embargo (-1) observations per splits"
        )

        return fig


def _n_splits(n_folds: int, n_test_folds: int) -> int:
    """Number of splits.

    Parameters
    ----------
    n_folds : int
        Number of folds.

    n_test_folds : int
        Number of test folds.

    Returns
    -------
    n_splits : int
        Number of splits
    """
    return int(math.comb(n_folds, n_test_folds))


def _n_test_paths(n_folds: int, n_test_folds: int) -> int:
    """Number of test paths that can be reconstructed from the train/test
    combinations

    Parameters
    ----------
    n_folds : int
        Number of folds.

    n_test_folds : int
        Number of test folds.

    Returns
    -------
    n_splits : int
        Number of test paths.
    """
    return (
        _n_splits(n_folds=n_folds, n_test_folds=n_test_folds) * n_test_folds // n_folds
    )


def _avg_train_size(n_observations: int, n_folds: int, n_test_folds: int) -> float:
    """Average number of observations contained in each training set.

    Parameters
    ----------
    n_observations : int
        Number of observations.

    n_folds : int
        Number of folds.

    n_test_folds : int
        Number of test folds.

    Returns
    -------
    avg_train_size : float
        Average number of observations contained in each training set.
    """
    return n_observations / n_folds * (n_folds - n_test_folds)


def optimal_folds_number(
    n_observations: int,
    target_train_size: int,
    target_n_test_paths: int,
    weight_train_size: float = 1,
    weight_n_test_paths: float = 1,
) -> tuple[int, int]:
    r"""Find the optimal number of folds (total folds and test folds) for a target
    training size and a target number of test paths.

    We find `x = n_folds` and `y = n_test_folds` that minimizes the below
    cost function of the relative distance from the two targets:

    .. math::
           cost(x,y) = w_{f} \times \lvert\frac{f(x,y)-f_{target}}{f_{target}}\rvert + w_{g} \times \lvert\frac{g(x,y)-g_{target}}{g_{target}}\rvert

    with :math:`w_{f}` and :math:`w_{g}` the weights assigned to the distance
    from each target and :math:`f(x,y)` and :math:`g(x,y)` the average training size
    and the number of test paths as a function of the number of total folds and test
    folds.

    This is a combinatorial problem with :math:`\frac{T\times(T-3)}{2}` combinations,
    with :math:`T` the number of observations.

    We reduce the search space by using the combinatorial symetry
    :math:`{n \choose k}={n \choose n-k}` and skipping cost computation above 1e5.

    Parameters
    ----------
    n_observations : int
        Number of observations.

    target_train_size : int
        The target number of observation in the training set.

    target_n_test_paths : int
        The target number of test paths (that can be reconstructed from the train/test
        combinations).

    weight_train_size : float, default=1
        The weight assigned to the distance from the target train size.
        The default value is 1.

    weight_n_test_paths : float, default=1
        The weight assigned to the distance from the target number of test paths.
        The default value is 1.

    Returns
    -------
    n_folds : int
        Optimal number of total folds.

    n_test_folds : int
        Optimal number of test folds.
    """

    def _cost(
        x: int,
        y: int,
    ) -> float:
        n_test_paths = _n_test_paths(n_folds=x, n_test_folds=y)
        avg_train_size = _avg_train_size(
            n_observations=n_observations, n_folds=x, n_test_folds=y
        )
        return (
            weight_n_test_paths
            * abs(n_test_paths - target_n_test_paths)
            / target_n_test_paths
            + weight_train_size
            * abs(avg_train_size - target_train_size)
            / target_train_size
        )

    costs = []
    res = []
    for n_folds in range(3, n_observations + 1):
        i = None
        for n_test_folds in range(2, n_folds):
            if i is None or n_folds - n_test_folds <= i:
                cost = _cost(x=n_folds, y=n_test_folds)
                costs.append(cost)
                res.append((n_folds, n_test_folds))
                if i is None and cost > 1e5:
                    i = n_test_folds

    j = np.argmin(costs)
    return res[j]



from skfolio.prior._base import BasePrior, PriorModel
from skfolio.prior._black_litterman import BlackLitterman
from skfolio.prior._empirical import EmpiricalPrior
from skfolio.prior._factor_model import (
    BaseLoadingMatrix,
    FactorModel,
    LoadingMatrixRegression,
)

__all__ = [
    "PriorModel",
    "BasePrior",
    "EmpiricalPrior",
    "BlackLitterman",
    "FactorModel",
    "BaseLoadingMatrix",
    "LoadingMatrixRegression",
]



"""Module that includes all Measures enums used across `skfolio`."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

from abc import abstractmethod
from enum import auto

from skfolio.utils.tools import AutoEnum


class BaseMeasure(AutoEnum):
    """Base Enum of measures"""

    def __repr__(self) -> str:
        """Enum representation for improved reading"""
        words = [
            (
                word.capitalize()
                if len(word) > 3
                else word.upper()
                if len(word) != 2
                else word.lower()
            )
            for word in self.value.split("_")
        ]
        if len(words[0]) in [3, 4] and words[0][-2:].lower() == "ar":
            s = list(words[0].upper())
            s[-2] = str(s[-2].lower())
            words[0] = "".join(s)
        string = " ".join(words).replace("Semi ", "Semi-")
        return string

    def __str__(self) -> str:
        return self.__repr__()

    @property
    @abstractmethod
    def is_perf(self):
        pass

    @property
    @abstractmethod
    def is_risk(self):
        pass

    @property
    @abstractmethod
    def is_ratio(self):
        pass

    @property
    def is_annualized(self) -> bool:
        return self.name[:10] == "ANNUALIZED"

    @property
    def annualized_measure(self):
        if self.is_annualized:
            raise ValueError(f"{self.name} is already an annualized measure")
        try:
            return getattr(self.__class__, f"ANNUALIZED_{self.name}")
        except AttributeError:
            raise AttributeError(
                f"{self.name} doesn't have a annualized version"
            ) from None

    @property
    def non_annualized_measure(self):
        if not self.is_annualized:
            raise ValueError(f"{self.name} is already a non-annualized measure")
        return getattr(self.__class__, self.name[11:])


class PerfMeasure(BaseMeasure):
    """Enumeration of performance measures

    Attributes
    ----------
    MEAN : str
        Mean

    ANNUALIZED_MEAN : str
       Annualized Mean
    """

    MEAN = auto()

    # Annualized measures
    ANNUALIZED_MEAN = auto()

    @property
    def is_perf(self) -> bool:
        return True

    @property
    def is_risk(self) -> bool:
        return False

    @property
    def is_ratio(self) -> bool:
        return False


class RiskMeasure(BaseMeasure):
    """Enumeration of risk measures

    Attributes
    ----------
    VARIANCE : str
        Variance.

    ANNUALIZED_VARIANCE : str
        Annualized Variance.

    SEMI_VARIANCE : str
        Semi-variance (Second Lower Partial Moment or Downside Variance).

    ANNUALIZED_SEMI_VARIANCE : str
        Annualized Semi-variance.

    STANDARD_DEVIATION : str
        Standard-deviation

    ANNUALIZED_STANDARD_DEVIATION : str
        Annualized Standard-deviation.

    SEMI_DEVIATION : str
        Semi-deviation.

    ANNUALIZED_SEMI_DEVIATION : str
        Annualized Semi-deviation.

    MEAN_ABSOLUTE_DEVIATION : str
        Mean Absolute Deviation.

    CVAR : str
        Conditional Value at Risk or Expected Shortfall.

    EVAR : str
        Entropic Value at Risk.

    WORST_REALIZATION : str
        Worst Realization (Worst Return).

    CDAR : str
        Conditional Drawdown at Risk.

    MAX_DRAWDOWN : str
        Maximum Drawdown.

    AVERAGE_DRAWDOWN : str
        Average Drawdown.

    EDAR : str
        Entropic Drawdown at Risk.

    FIRST_LOWER_PARTIAL_MOMENT : str
        First Lower Partial Moment.

    ULCER_INDEX : str
        Ulcer Index.

    GINI_MEAN_DIFFERENCE : str
        Gini Mean Difference.
    """

    VARIANCE = auto()
    ANNUALIZED_VARIANCE = auto()
    SEMI_VARIANCE = auto()
    ANNUALIZED_SEMI_VARIANCE = auto()
    STANDARD_DEVIATION = auto()
    ANNUALIZED_STANDARD_DEVIATION = auto()
    SEMI_DEVIATION = auto()
    ANNUALIZED_SEMI_DEVIATION = auto()
    MEAN_ABSOLUTE_DEVIATION = auto()
    CVAR = auto()
    EVAR = auto()
    WORST_REALIZATION = auto()
    CDAR = auto()
    MAX_DRAWDOWN = auto()
    AVERAGE_DRAWDOWN = auto()
    EDAR = auto()
    FIRST_LOWER_PARTIAL_MOMENT = auto()
    ULCER_INDEX = auto()
    GINI_MEAN_DIFFERENCE = auto()

    @property
    def is_perf(self) -> bool:
        return False

    @property
    def is_risk(self) -> bool:
        return True

    @property
    def is_ratio(self) -> bool:
        return False


class ExtraRiskMeasure(BaseMeasure):
    """Enumeration of other risk measures not used in convex optimization

    Attributes
    ----------
    VALUE_AT_RISK : str
        Value at Risk (VaR).

    DRAWDOWN_AT_RISK : str
        Drawdown at Risk.

    ENTROPIC_RISK_MEASURE : str
        Entropic Risk Measure.

    FOURTH_CENTRAL_MOMENT : str
        Fourth Central Moment.

    FOURTH_LOWER_PARTIAL_MOMENT : str
         Fourth Lower Central Moment.

    SKEW : str
        Skew.

    KURTOSIS : str
        Kurtosis.
    """

    VALUE_AT_RISK = auto()
    DRAWDOWN_AT_RISK = auto()
    ENTROPIC_RISK_MEASURE = auto()
    FOURTH_CENTRAL_MOMENT = auto()
    FOURTH_LOWER_PARTIAL_MOMENT = auto()
    SKEW = auto()
    KURTOSIS = auto()

    @property
    def is_perf(self) -> bool:
        return False

    @property
    def is_risk(self) -> bool:
        return True

    @property
    def is_ratio(self) -> bool:
        return False


class RatioMeasure(BaseMeasure):
    """Enumeration of ratio measures

    Attributes
    ----------
    SHARPE_RATIO : str
        Ratio of the excess Mean divided by the Standard-deviation.

    ANNUALIZED_SHARPE_RATIO : str
        Annualized Sharpe ratio.

    SORTINO_RATIO : str
         Ratio of the excess Mean divided by the Semi standard-deviation.

    ANNUALIZED_SORTINO_RATIO : str
        Annualized Sortino ratio.

    MEAN_ABSOLUTE_DEVIATION_RATIO : str
        Ratio of the excess Mean divided by the Mean Absolute Deviation.

    FIRST_LOWER_PARTIAL_MOMENT_RATIO : str
        Ratio of the excess Mean divided by the First Lower Partial Moment.

    VALUE_AT_RISK_RATIO : str
         Ratio of the excess Mean divided by the Value at Risk.

    CVAR_RATIO : str
         Ratio of the excess Mean divided by the Conditional Value at Risk.

    ENTROPIC_RISK_MEASURE_RATIO : str
         Ratio of the excess Mean divided by the Entropic Risk Measure.

    EVAR_RATIO : str
         Ratio of the excess Mean divided by the Entropic Value at Risk.

    WORST_REALIZATION_RATIO : str
         Ratio of the excess Mean divided by the Worst Realization.

    DRAWDOWN_AT_RISK_RATIO : str
         Ratio of the excess Mean divided by the Drawdown at Risk.

    CDAR_RATIO : str
         Ratio of the excess Mean divided by the Conditional Drawdown at Risk.

    CALMAR_RATIO : str
         Ratio of the excess Mean divided by the Maximum Drawdown.

    AVERAGE_DRAWDOWN_RATIO : str
         Ratio of the excess Mean divided by the Average Drawdown.

    EDAR_RATIO : str
         Ratio of the excess Mean divided by the Entropic Drawdown at Risk.

    ULCER_INDEX_RATIO : str
         Ratio of the excess Mean divided by the Ulcer Index.

    GINI_MEAN_DIFFERENCE_RATIO : str
         Ratio of the excess Mean divided by the Gini Mean Difference.
    """

    SHARPE_RATIO = auto()
    ANNUALIZED_SHARPE_RATIO = auto()
    SORTINO_RATIO = auto()
    ANNUALIZED_SORTINO_RATIO = auto()
    MEAN_ABSOLUTE_DEVIATION_RATIO = auto()
    FIRST_LOWER_PARTIAL_MOMENT_RATIO = auto()
    VALUE_AT_RISK_RATIO = auto()
    CVAR_RATIO = auto()
    ENTROPIC_RISK_MEASURE_RATIO = auto()
    EVAR_RATIO = auto()
    WORST_REALIZATION_RATIO = auto()
    DRAWDOWN_AT_RISK_RATIO = auto()
    CDAR_RATIO = auto()
    CALMAR_RATIO = auto()
    AVERAGE_DRAWDOWN_RATIO = auto()
    EDAR_RATIO = auto()
    ULCER_INDEX_RATIO = auto()
    GINI_MEAN_DIFFERENCE_RATIO = auto()

    @property
    def is_perf(self) -> bool:
        return False

    @property
    def is_risk(self) -> bool:
        return False

    @property
    def is_ratio(self) -> bool:
        return True

    @property
    def linked_risk_measure(self) -> RiskMeasure | ExtraRiskMeasure:
        match self:
            case RatioMeasure.SHARPE_RATIO:
                return RiskMeasure.STANDARD_DEVIATION
            case RatioMeasure.ANNUALIZED_SHARPE_RATIO:
                return RiskMeasure.ANNUALIZED_STANDARD_DEVIATION
            case RatioMeasure.SORTINO_RATIO:
                return RiskMeasure.SEMI_DEVIATION
            case RatioMeasure.ANNUALIZED_SORTINO_RATIO:
                return RiskMeasure.ANNUALIZED_SEMI_DEVIATION
            case RatioMeasure.CALMAR_RATIO:
                return RiskMeasure.MAX_DRAWDOWN
            case _:
                risk_measure = str(self.value).replace("_ratio", "")
                try:
                    return RiskMeasure(risk_measure)
                except ValueError:
                    return ExtraRiskMeasure(risk_measure)



"""Base Expected returns estimators."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt
import sklearn.base as skb


class BaseMu(skb.BaseEstimator, ABC):
    """Base class for all expected returns estimators in skfolio.

    Notes
    -----
    All estimators should specify all the parameters that can be set
    at the class level in their ``__init__`` as explicit keyword
    arguments (no ``*args`` or ``**kwargs``).
    """

    mu_: np.ndarray

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def fit(self, X: npt.ArrayLike, y=None):
        pass
